import json
import logging
import time
from datetime import datetime
from typing import Any, Callable, Dict, Generator, List, Optional, Tuple, Union
from urllib.parse import urlparse

import gitlab
import requests
from gitlab import Gitlab
from gitlab.base import RESTObject
from gitlab.exceptions import GitlabError
from gitlab.v4.objects import Group, Project, ProjectBranch, ProjectCommit
from gitlab.v4.objects import User as GitlabUser

from jf_ingest import logging_helper
from jf_ingest.config import GitAuthConfig, GitLabAuthConfig
from jf_ingest.constants import Constants
from jf_ingest.graphql_utils import GQL_PAGE_INFO_BLOCK
from jf_ingest.jf_git.exceptions import (
    GitAuthenticationException,
    GqlPageTimeoutException,
)
from jf_ingest.jf_git.standardized_models import (
    StandardizedOrganization,
    StandardizedPullRequest,
    StandardizedRepository,
)
from jf_ingest.utils import (
    GITLAB_STATUSES_TO_RETRY,
    RetryLimitExceeded,
    hash_filename,
    parse_gitlab_api_version,
    retry_for_status,
    retry_session,
)

logger = logging.getLogger(__name__)


class GitlabClient:
    # NOTE: We currently have some functions in this class that access GraphQL.
    # When trying to implement that, we ran into some authentication issues and limitations that moved us toward
    #   using the rest client instead.
    # Functions that make use of GraphQL are suffixed with _gql.
    # If we want to keep them, we'd ideally eventually move them to their own subclass of the client.
    GITLAB_GQL_USER_FRAGMENT = "... on User {login, id: databaseId, email, name, url}"
    GITLAB_GQL_USER_NODES = 'id, name, username, webUrl, publicEmail'
    GITLAB_GQL_SHORT_REPO_NODES = 'id, name, webUrl'

    def __init__(self, auth_config: GitAuthConfig, **kwargs):
        """Gitlab Client, used as a wrapper for getting raw data from the API.
        This client will get data mainly from the GraphQL API endpoints, although
        it will use the REST API endpoints for a small amount of functions as well.

        Args:
            auth_config (GitAuthConfig): A valid GitAuthConfiguration object. MUST BE A GitLabAuthConfig object, NOT A GitAuthConfig OBJECT!!
            kwargs: kwargs are used to pass arguments to the inner Session object, if no session object is provided as part of the GitAuthConfig
        """
        if type(auth_config) != GitLabAuthConfig:
            raise Exception(
                f'Unknown Auth Config provided to GitlabClient. GitAuthConfig type is expected to be a GitLabAuthConfig type, but received {type(auth_config)}'
            )
        self.company_slug: str = auth_config.company_slug
        self.rest_api_url: Optional[str] = auth_config.base_url
        self.gql_base_url: str = f'{auth_config.base_url}/api/graphql'
        if session := auth_config.session:
            self.session: requests.Session = session
        else:
            self.session = retry_session(**kwargs)
            self.session.headers.update(
                {
                    'Authorization': f'Bearer {auth_config.token}',
                    'Content-Type': 'application/json',
                    'User-Agent': f'{Constants.JELLYFISH_USER_AGENT} ({requests.utils.default_user_agent()})',
                }
            )
        self.session.verify = auth_config.verify
        self.client: Gitlab = Gitlab(
            url=self.rest_api_url,
            session=self.session,
            ssl_verify=auth_config.verify,
            keep_base_url=auth_config.keep_base_url,
        )
        try:
            self.compatibility_config = self.get_api_compatibility_config()
        except RetryLimitExceeded:
            logging_helper.send_to_agent_log_file(
                'Max number of retries reached when attempting to get Server Version. '
                'I think it\'s safe to assume we are having authentication problems. '
                'Raising GitAuthenticationException.',
                level=logging.WARNING,
            )
            raise GitAuthenticationException(
                'Max number of retries reached when attempting to get the Gitlab Server version from API endpoint /api/v4/metadata. '
                'This likely indicates that we are having authentication problems.'
            )

    @property
    def supports_get_all_organizations(self):
        # listing all organizations is available for private Gitlab instances ONLY
        # otherwise, we will be querying the _entire_ public repository
        if not self.rest_api_url:
            return False

        hostname = urlparse(self.rest_api_url).hostname
        return not hostname.endswith(Constants.GITLAB_PUBLIC_HOSTNAME)

    def get_api_version(self) -> str:
        try:
            resp = retry_for_status(self.client.http_get, '/metadata')
        except Exception as e:
            logger.warning(
                f'Unable to fetch server version via /metadata. Checking /version... Exception: {e}'
            )
            resp = retry_for_status(self.client.http_get, '/version')

        return str(resp['version'])

    def get_api_compatibility_config(self) -> Dict:
        """
        Some versions of the api may be missing functionality. This function is designed to help us work around those.

        Returns:
            compatibility_mapping Dict: A collection of keys and values to help the client handle version differences.
        """
        compatibility_config = {
            'gql_skip_pr_closed_at': False,
            'gql_skip_commit_committed_date': False,
        }

        # Get version stripped of any tags or other letters. This should align roughly with semver
        api_version: str = self.get_api_version()

        if api_version:
            major_version = None
            minor_version = None

            try:
                api_version = parse_gitlab_api_version(api_version)
                version_split = api_version.split('.')
                major_version = int(version_split[0])

                if len(version_split) > 1:
                    minor_version = int(version_split[1])
            except ValueError:
                logger.warning(
                    ':WARN: Unable to determine GitLab server version. Enabling all compatability configuration.'
                )

            # API versions prior to 17.4 do not have closedAt available on the mergedRequest. We can still fetch the rest
            # of the data we want via GQL, but that field needs to be populated via the api for customers on older versions.
            if (not major_version or major_version < 17) or (
                major_version == 17 and (not minor_version or minor_version < 4)
            ):
                compatibility_config['gql_skip_pr_closed_at'] = True

            if not major_version or major_version < 16:
                compatibility_config['gql_skip_commit_committed_date'] = True

        return compatibility_config

    def get_organization_name_full_path_and_url(self, login: str) -> Tuple[str, str, str]:
        """In Jellyfish Land, the JFGithubOrganization is the normalization of Github Organizations,
        AzureDevops Organizations, Bitbucket Projects, and Gitlab Groups. The login field is the unique
        key. For Gitlab Groups, we set the login to the be the Group ID, which is a numeric value.
        The GraphQL Group Queries accept a "fullPath" argument, and NOT the Group ID. If we only have
        the GroupID (set by the login value), then this helper function can be used to translate the
        GroupID to a Full Path.
        NOTE: For performance reasons, we should probably graph the FullPath when we query GraphQL for
        Groups in general, and then cache those values. We should NOT call this function everytime,
        because it could have performance implications

        Args:
            login (str): The JFGithubOrganization login, which is the Group ID in Gitlab land
        Returns:
            name, full_path, url (str, str, str): The name, Full Path, and url for this gitlab Group
        """
        group_url = f'{self.rest_api_url}/api/v4/groups/{login}?with_projects=False'
        response: requests.Response = retry_for_status(
            self.session.get, url=group_url, statuses_to_retry=GITLAB_STATUSES_TO_RETRY
        )
        response.raise_for_status()
        response_json = response.json()
        return (
            str(response_json['name']),
            str(response_json['full_path']),
            str(response_json['web_url']),
        )

    def get_raw_result_gql(self, query_body: str, max_attempts: int = 5) -> Dict:
        """Gets the raw results from a Graphql Query.

        Args:
            query_body (str): A query body to hit GQL with
            max_attempts (int, optional): The number of retries we should make when we specifically run into GQL Rate limiting.
                This value is important if the GQL endpoint doesn't give us (or gives us a malformed) rate limit header.
                Defaults to 5.

        Raises:
            GqlRateLimitExceededException: A custom exception if we run into GQL rate limiting (200 response) and we run out of attempts (based on max_attempts)
            RetryLimitExceeded: A custom exception thrown by retry_for_status if we exceed our retries (based on max_attempts)
            Exception: Any other random exception we encounter, although the big rate limiting use cases are generally covered

        Returns:
            dict: A raw dictionary result from GQL
        """
        response: requests.Response = retry_for_status(
            self.session.post,
            url=self.gql_base_url,
            json={'query': query_body},
            max_retries_for_retry_for_status=max_attempts,
        )
        json_str = response.content.decode()
        json_data: Dict = json.loads(json_str)
        if error_list_dict := json_data.get('errors'):
            error_message = ','.join([error_dict.get('message') for error_dict in error_list_dict])

            # If we received a "Timeout on..." response, we should raise this specific exception so we can attempt with
            # a smaller page size.
            if response.status_code == 200 and error_message.startswith('Timeout on'):
                logging_helper.send_to_agent_log_file(
                    f'Timeout detected when querying for data. Received error: {error_message}',
                    level=logging.WARNING,
                )
                raise GqlPageTimeoutException({error_message})
            else:
                raise Exception(
                    f'An error occurred when attempting to query GraphQL (response code {response.status_code}): {error_message}'
                )

        return json_data

    def page_results_gql(
        self,
        query_body: str,
        path_to_page_info: str,
        cursor: Optional[str] = None,
        page_size: int = 100,
    ) -> Generator[Dict, None, None]:
        """
        This is a helper function for paging results from GraphQL.
        It expects a query body that has "%s" for both `first` (page_size) and `after` (cursor).
        THE ORDER IS IMPORTANT, `first` (page_size)` MUST COME BEFORE `after` (cursor) IN THE QUERY BODY.
        BOTH MUST BE PRESENT.
        QUERY BODIES MUST INCLUDE VALID PAGE INFO (see GQL_PAGE_INFO_BLOCK).

        Args:
            query_body (str): The query body to hit GraphQL with
            path_to_page_info (str): A string of period separated words that lead
                to the part of the query that we are trying to page. Example: data.organization.userQuery
            cursor (str, optional): This argument is used to recursively to page.
                The cursor is provided by GitLab (endCursor) and is used to denote the next item to start from.
                Defaults to 'null'. LEAVE AS 'null' OR `None` if fetching first page.
            page_size (int): Page size to use in query. This can be altered by our backoff if we run into timeouts.
                Defaults to 100.

        Yields:
            Generator[dict, None, None]: This function yields each item from all the pages paged, item by item
        """
        hasNextPage = True
        if not cursor:
            cursor = 'null'
        else:
            cursor = f'"{cursor}"'

        while hasNextPage:
            try:
                # Fetch results
                result = self.get_raw_result_gql(query_body=query_body % (page_size, cursor))

                yield result

                # Check if there are additional pages.
                page_info = self.get_path_value(path_to_page_info + '.pageInfo', result) or {}
                _cursor = page_info.get('endCursor', None)
                hasNextPage = page_info.get('hasNextPage', False) and _cursor

                # Cursor must be wrapped in quotes.
                cursor = f'"{_cursor}"'
            except GqlPageTimeoutException as e:
                # Sometimes we time out on validation of a query or if some of the nested objects in a query timeout.
                # This can be helped by reducing the pagination size for the remainder of the queries.
                # GitLab reuses cursors even if the page size or payload selection changes.
                # In the future, we can choose to skip selecting nested fields if they consistently time out.
                if page_size <= 1:
                    # TODO: We should probably track this with JellyCounter
                    logging_helper.send_to_agent_log_file(
                        f"Unable to successfully query for data, due to excessive timeouts. Breaking out of paging loop",
                        level=logging.ERROR,
                    )
                    break
                else:
                    new_page_size = max(1, page_size - 10)
                    logging_helper.send_to_agent_log_file(
                        f"Received timeout with page size {page_size}. Reducing to {new_page_size}...",
                        level=logging.WARNING,
                    )
                    page_size = new_page_size
                    # Wait 30 seconds here to possibly help with any hung up resources
                    time.sleep(30)
            except RetryLimitExceeded as e:
                logging_helper.send_to_agent_log_file(
                    f"Exceeded retry limit when fetching data. Breaking out of paging loop. Exception: {e}",
                    level=logging.ERROR,
                )
                break

    def page_results(
        self, gitlab_function: Callable, starting_page: int = 1, page_size: int = 100, **kwargs
    ) -> Generator[Project | RESTObject, None, None]:
        """
        This is a helper function for paging results from a gitlab function. This is done to prevent
        ChunkedEncodingErrors from being uncaught when using the `iterator=True` kwarg.

        Args:
            gitlab_function (Callable): A Gitlab library function to be called with pagination.
            starting_page (int): Page to start from when returning items. 1 is the first page. Defaults to 1.
            page_size (int): Page size to use in kwargs of function. Defaults to 100.

        Yields:
            Generator[dict, None, None]: This function yields each item from all the pages paged, item by item
        """
        kwargs['page'] = starting_page
        kwargs['per_page'] = page_size
        while True:
            results = retry_for_status(gitlab_function, statuses_to_retry=GITLAB_STATUSES_TO_RETRY, **kwargs)  # type: ignore
            if not results:
                break
            for result in results:
                yield result
            kwargs['page'] += 1

    def get_path_value(self, path_value: str, data_dict: Optional[dict] = None) -> Optional[Any]:
        """
        This function takes a path value separated by periods to determine nesting in a dict. This will be primarily
        used to safely access response pages from GQL. We log a warning if we're provided an empty dict or None.

        Args:
            path_value (str): A period-delineated string of keys to recursively traverse and then return the last key.
                If the key is not present, we will return an empty dictionary.
                Example: data.project.mergeRequests.nodes
            data_dict (Optional[dict]): Dictionary to traverse. If not provided

        Returns:
            Optional[Any]: If there is a value for the given key path, it's returned.
                Defaults to an empty dict.
        """
        if not data_dict:
            logging_helper.send_to_agent_log_file(
                f"Null or empty dict provided for path value {path_value}.",
                level=logging.WARNING,
            )
            return {}

        result = data_dict

        path_tokens = path_value.split('.')
        for token in path_tokens:
            result = result.get(token, {})

        return result

    def get_teams(self, *args, **kwargs) -> list:
        """
        This function is to align with other clients.
        GitLab does not have a concept of teams past groups, which we use as organizations.
        This will return an empty list, regardless of arguments.
        """
        return []

    def get_repos(
        self, jf_org: StandardizedOrganization, only_private: bool = False
    ) -> Generator[Project | RESTObject, None, None]:
        try:
            group = retry_for_status(
                self.client.groups.get, jf_org.login, statuses_to_retry=GITLAB_STATUSES_TO_RETRY
            )

            kwargs_dict: dict[str, Any] = {'include_subgroups': True}
            if only_private:
                kwargs_dict['visibility'] = 'private'

            for repo in self.page_results(group.projects.list, **kwargs_dict):
                yield repo
        except (requests.exceptions.ConnectionError, GitlabError, RetryLimitExceeded) as e:
            logger.warning(
                f"Failed to pull repos for group {jf_org.login}. Skipping. Exception: {e}"
            )

    def get_repos_count(self, login: str, only_private: Optional[bool] = False) -> int:
        url = f"{self.rest_api_url}/api/v4/groups/{login}/projects"
        params: dict[str, Any] = {'per_page': 1}

        if only_private:
            params['visibility'] = 'private'

        try:
            response: requests.Response = retry_for_status(
                self.session.get, url=url, params=params, statuses_to_retry=GITLAB_STATUSES_TO_RETRY
            )
            return int(response.headers.get('X-Total', 0))
        except (requests.exceptions.ConnectionError, GitlabError, RetryLimitExceeded) as e:
            logger.warning(f"Failed to get repo count for {login}. Skipping. Exception: {e}")
            return 0

    def get_commits(
        self,
        jf_repo: StandardizedRepository,
        since: Optional[datetime] = None,
        until: Optional[datetime] = None,
        branch_name: Optional[str] = None,
    ) -> Generator[ProjectCommit | RESTObject, None, None]:
        project_id = jf_repo.id
        project = retry_for_status(
            self.client.projects.get, project_id, statuses_to_retry=GITLAB_STATUSES_TO_RETRY
        )

        kwargs_dict: dict[str, Any] = {}
        if branch_name:
            kwargs_dict['ref_name'] = branch_name
        if since:
            kwargs_dict['since'] = since
        if until:
            kwargs_dict['until'] = until

        try:
            for commit in self.page_results(project.commits.list, **kwargs_dict):
                yield commit

        except (requests.exceptions.ConnectionError, GitlabError, RetryLimitExceeded) as e:
            logger.warning(
                f"Failed to pull commits for repo {jf_repo.name}. Skipping. Exception: {e}"
            )

    def get_branches_for_repo(
        self, jf_repo: StandardizedRepository, branch_name: Optional[str] = None
    ) -> Generator[ProjectBranch | RESTObject, None, None]:
        """
        Attempts to pull branches for a given repository.

        Args:
            standardized_repo (StandardizedRepository): A standardized repo, which is used to pull repositories.
            branch_name (Optional[str]): Name of the branch to pull. If this is provided, only attempt to pull this
                branch. Otherwise, attempts to pull all branches.

        Yields:
            ProjectBranch | RESTObject: Objects generated by the GitLab client.
        """
        project_id = jf_repo.id
        project = retry_for_status(
            self.client.projects.get, project_id, statuses_to_retry=GITLAB_STATUSES_TO_RETRY
        )
        branches_list: List[ProjectBranch | RESTObject] = []

        kwargs_dict: dict[str, Any] = {}
        if branch_name:
            try:
                branches_list.append(project.branches.get(branch_name))
            except gitlab.exceptions.GitlabGetError:
                try:
                    # Per legacy code (scc 1/9/2023), we may hit this for very old GitLab servers (<11.9).
                    # In this case, we hit the branch list endpoint and use the given branch name as the search term.
                    logger.info(
                        f"Failed to fetch branch '{branch_name}' for '{jf_repo.name}' via get, falling back to search."
                    )

                    # We technically could use regex here to reduce extra hits, but we're not guaranteed to have regex
                    # on very old versions of GitLab's api. Limit this to 10 per page, likely to have a hit there.
                    kwargs_dict['search'] = branch_name
                    branches = self.page_results(project.branches.list, page_size=10, **kwargs_dict)

                    # If we're given a branch name, we only expect one branch back. We *may* get more than one, so just
                    # return the first we get that completely matches the branch name.
                    found_branch = None
                    for branch in branches:
                        if branch.name == branch_name:
                            found_branch = branch
                            branches_list.append(branch)
                            break

                    if not found_branch:
                        logger.warning(
                            f"Failed to find branch '{branch_name}' for '{jf_repo.name}', skipping..."
                        )

                except gitlab.exceptions.GitlabListError:
                    # We believe Gitlab does this to mean "0 results".
                    logger.warning(
                        f"Failed to find branch '{branch_name}' for '{jf_repo.name}', skipping..."
                    )
        else:
            try:
                yield from retry_for_status(project.branches.list, statuses_to_retry=GITLAB_STATUSES_TO_RETRY, **kwargs_dict)  # type: ignore
            except gitlab.exceptions.GitlabListError:
                # We believe Gitlab does this to mean "0 results".
                logger.warning(f"Failed to fetch branches for '{jf_repo.name}', skipping...")

        for branch in branches_list:
            yield branch

    def get_organizations_rest_api(self) -> Generator[Group | RESTObject, None, None]:
        kwargs_dict: dict[str, Any] = {'query_data': {'all_available': True}}
        for group in self.page_results(self.client.groups.list, **kwargs_dict):
            if getattr(group, 'parent_id', None) is None:
                yield group

    def get_organizations_gql(
        self, page_size: int = 100, sort_key: str = 'id_asc'
    ) -> Generator[Dict, None, None]:
        query_body = f"""
        {{
            groupsQuery: groups(first: %s, sort: "{sort_key}", after: %s){{
                {GQL_PAGE_INFO_BLOCK}
                groups: nodes {{
                    groupIdStr: id
                    name
                    fullPath
                    webUrl
                }}
            }}
        }}
        """
        for page in self.page_results_gql(
            query_body=query_body, path_to_page_info='data.groupsQuery', page_size=page_size
        ):
            groups = self.get_path_value('data.groupsQuery.groups', page) or []
            for group in groups:
                yield group

    def get_repos_gql(
        self, group_full_path: str, page_size: int = 100
    ) -> Generator[Dict, None, None]:
        query_body = f"""
        {{
            group(fullPath: "{group_full_path}") {{
                projectsQuery: projects(first: %s, after: %s) {{
                    {GQL_PAGE_INFO_BLOCK}
                    projects: nodes {{
                        ... on Project {{
                            name,
                            webUrl,
                            description,
                            isForked,
                            repository {{
                                ... on Repository {{
                                    defaultBranchName: rootRef
                                }}
                            }}
                        }}
                    }}
                }}
            }}
        }}
        """
        for page in self.page_results_gql(
            query_body=query_body, path_to_page_info='data.group.projectsQuery', page_size=page_size
        ):
            projects = self.get_path_value('data.group.projectsQuery.projects', page) or []
            for project in projects:
                yield project

    def get_users(self, group_id: str) -> Generator[Union[RESTObject, GitlabUser], None, None]:
        """
        Gets all users for a given Gitlab group (aka organization)

        Args:
            group_id: ID of the group (organization) to get users for

        Returns:
            Generator[Union[RESTObject, GitlabUser], None, None]: Generator yielding one user object at a time
        """
        try:
            group = retry_for_status(
                self.client.groups.get, group_id, statuses_to_retry=GITLAB_STATUSES_TO_RETRY
            )
            for user in self.page_results(group.members.list):
                yield user

        except Exception as e:
            logging_helper.send_to_agent_log_file(
                f'When attempting to get user data we ran in the following exception: {e}. '
                f'Giving up on pulling users for group {group_id} and moving on',
                level=logging.WARNING,
            )

    def get_prs(
        self,
        project_full_path: str,
        start_cursor: Optional[Any] = None,
        start_window: Optional[datetime] = None,
        end_window: Optional[datetime] = None,
        pull_files_for_pr: Optional[bool] = False,
        hash_files_for_pr: Optional[bool] = False,
        page_size: int = 100,
    ) -> Generator[Dict, None, None]:
        """
        Gets all pull requests for a given Gitlab project (aka repository) using GraphQL

        Args:
            project_full_path (str): Full path of the project (repository) to get pull requests for
            start_cursor (Optional[str]): A cursor string to start from when fetching prs, defaults to None
            start_window (Optional[datetime]): Filter prs to those updated after this date. defaults to None
            end_window (Optional[datetime]): Filter prs to those updated before this date. defaults to None
            pull_files_for_pr (Optional[bool]): Used to determine if we should pull file data with prs, defaults to False
            hash_files_for_pr (Optional[bool]): Used to determine if pulled filenames should be hashed, defaults to False
                Only applied to filenames, so will have no effect if pull_files_for_pr is False
            page_size (int): Page size for the API call, defaults to 100

        Returns:
            Generator[Dict, None, None]: Generator yielding one pull request dict at a time
        """
        query_body = f"""
        {{
            project(fullPath: "{project_full_path}") {{
                mergeRequests(first: %s, after: %s, sort: UPDATED_DESC{f', updatedAfter: "{start_window.isoformat()}"' if start_window else ''}{f', updatedBefore: "{end_window.isoformat()}"' if end_window else ''}) {{
                    {GQL_PAGE_INFO_BLOCK}
                    nodes {{
                        id
                        iid
                        {'closedAt' if not self.compatibility_config['gql_skip_pr_closed_at'] else ''}
                        updatedAt
                        mergedAt
                        createdAt
                        title
                        description
                        webUrl
                        sourceBranch
                        targetBranch
                        mergeCommitSha
                        diffStatsSummary {{ additions, deletions, fileCount }}
                        {'diffStats { additions, deletions, path }' if pull_files_for_pr else ''}
                        sourceProject {{ {self.GITLAB_GQL_SHORT_REPO_NODES} }}
                        targetProject {{ {self.GITLAB_GQL_SHORT_REPO_NODES} }}
                        author {{ {self.GITLAB_GQL_USER_NODES} }}
                        mergeUser {{ {self.GITLAB_GQL_USER_NODES} }}
                        {self._get_pr_commits_query_block(enable_paging=False)}
                        {self._get_pr_comments_query_block(enable_paging=False)}
                        {self._get_pr_approvals_query_block(enable_paging=False)}
                    }}
                }}
            }}
        }}
        """

        for page in self.page_results_gql(
            query_body=query_body,
            path_to_page_info='data.project.mergeRequests',
            cursor=start_cursor,
            page_size=page_size,
        ):
            pr_nodes = self.get_path_value('data.project.mergeRequests.nodes', page) or []
            for pr_node in pr_nodes:
                pr_iid = pr_node['iid']

                # Fetch any commits not fetched on initial page
                if self.get_path_value('commits.pageInfo.hasNextPage', pr_node):
                    commits = [
                        commit
                        for commit in self.get_pr_commits(
                            project_full_path=project_full_path,
                            pr_iid=pr_iid,
                        )
                    ]
                    pr_node['commits']['nodes'] = commits

                # Fetch any comments not fetched on initial page
                if self.get_path_value('notes.pageInfo.hasNextPage', pr_node):
                    comments = [
                        comment
                        for comment in self.get_pr_comments(
                            project_full_path=project_full_path,
                            pr_iid=pr_iid,
                        )
                    ]
                    pr_node['notes']['nodes'] = comments

                # Fetch any approvals not fetched on initial page
                if self.get_path_value('approvedBy.pageInfo.hasNextPage', pr_node):
                    approvals = [
                        approval
                        for approval in self.get_pr_approvals(
                            project_full_path=project_full_path,
                            pr_iid=pr_iid,
                        )
                    ]
                    pr_node['approvedBy']['nodes'] = approvals

                # Hash filenames if enabled and we have file paths to hash
                if pull_files_for_pr and hash_files_for_pr:
                    hashed_files = []
                    for unhashed_file in pr_node.get('diffStats', []):
                        unhashed_file['path'] = hash_filename(unhashed_file['path'])
                        hashed_files.append(unhashed_file)
                    pr_node['diffStats'] = hashed_files

                yield pr_node

    def get_pr_commits(
        self,
        project_full_path: str,
        pr_iid: str,
        page_size: int = 20,
    ) -> Generator[dict, None, None]:
        query_body = f"""   
        {{         
            project(fullPath: "{project_full_path}") {{
                mergeRequest(iid: "{pr_iid}") {{
                    {self._get_pr_commits_query_block(enable_paging=True)}
                }}
            }}
        }}
        """
        for page in self.page_results_gql(
            query_body=query_body,
            path_to_page_info='data.project.mergeRequest.commits',
            page_size=page_size,
        ):
            nodes = self.get_path_value('data.project.mergeRequest.commits.nodes', page) or []
            for node in nodes:
                yield node

    def get_pr_commits_rest_api(
        self, standardized_repo: StandardizedRepository, api_pr: dict
    ) -> Generator[dict, None, None]:
        commits_path = (
            f'/projects/{standardized_repo.id}/merge_requests/{int(api_pr["iid"])}/commits'
        )
        return retry_for_status(self.client.http_list, commits_path, iterator=True)  # type: ignore

    def get_pr_comments(
        self,
        project_full_path: str,
        pr_iid: str,
        page_size: int = 20,
    ) -> Generator[dict, None, None]:
        query_body = f"""
        {{      
            project(fullPath: "{project_full_path}") {{
                mergeRequest(iid: "{pr_iid}") {{
                    {self._get_pr_comments_query_block(enable_paging=True)}
                }}
            }}
        }}
        """
        for page in self.page_results_gql(
            query_body=query_body,
            path_to_page_info='data.project.mergeRequest.notes',
            page_size=page_size,
        ):
            nodes = self.get_path_value('data.project.mergeRequest.notes.nodes', page) or []
            for node in nodes:
                yield node

    def get_pr_approvals(
        self,
        project_full_path: str,
        pr_iid: str,
        page_size: int = 20,
    ) -> Generator[dict, None, None]:
        query_body = f"""
        {{       
            project(fullPath: "{project_full_path}") {{
                mergeRequest(iid: "{pr_iid}") {{
                    {self._get_pr_approvals_query_block(enable_paging=True)}
                }}
            }}
        }}
        """
        for page in self.page_results_gql(
            query_body=query_body,
            path_to_page_info='data.project.mergeRequest.approvedBy',
            page_size=page_size,
        ):
            nodes = self.get_path_value('data.project.mergeRequest.approvedBy.nodes', page) or []
            for node in nodes:
                yield node

    def _get_pr_commits_query_block(self, enable_paging: bool = False):
        return f"""
            commits(first: {'%s, after: %s' if enable_paging else '5'}) {{
                {GQL_PAGE_INFO_BLOCK}
                nodes {{
                    id
                    sha
                    webUrl
                    message
                    {'committedDate' if not self.compatibility_config['gql_skip_commit_committed_date'] else ''}
                    authoredDate
                    author {{ {self.GITLAB_GQL_USER_NODES} }}
                }}
            }}
        """

    def _get_pr_comments_query_block(self, enable_paging: bool = False):
        return f"""
            notes(first: {'%s, after: %s' if enable_paging else '5'}) {{
                {GQL_PAGE_INFO_BLOCK}
                nodes {{
                    id
                    body
                    createdAt
                    system
                    author {{ {self.GITLAB_GQL_USER_NODES} }}
                }}
            }}
        """

    def _get_pr_approvals_query_block(self, enable_paging: bool = False):
        return f"""
            approvedBy(first: {'%s, after: %s' if enable_paging else '5'}) {{
                {GQL_PAGE_INFO_BLOCK}
                nodes {{ {self.GITLAB_GQL_USER_NODES} }}
            }}
        """



from dataclasses import dataclass, field
from datetime import date, datetime
from typing import ClassVar, List, Optional, TypeAlias, TypeVar

from jf_ingest.data_manifests.manifest_base import ManifestBase


# Separating the model from the adapter behavior as a base class to make the data more transportable and reusable
@dataclass
class JiraProjectManifestBase(ManifestBase):
    project_id: Optional[str] = field(default=None)
    project_key: Optional[str] = field(default=None)
    issues_count: Optional[int] = field(default=None)
    version_count: Optional[int] = field(default=None)
    # This reflects if a client has intentionally excluded
    # a project from being ingested
    excluded: bool = field(default=False)
    # Pull from date
    pull_from: Optional[date] = field(default=None)
    # Latest issue updated date
    last_issue_updated_date: Optional[datetime] = field(default=None)
    # Allocation Status
    classification: Optional[int] = field(default=None)
    # Human readable allocation status.
    # Values pulled from JiraProjectClassification.CLASSIFICATION_CHOICES
    classification_str: Optional[str] = field(default=None)


IJiraProjectManifest = TypeVar('IJiraProjectManifest', bound='JiraProjectManifestBase')
IJiraProjectManifestList: TypeAlias = list[IJiraProjectManifest]


# Separating the model from the adapter behavior as a base class to make the data more transportable and reusable
@dataclass
class JiraDataManifestBase(ManifestBase):
    # Pull from date
    pull_from: Optional[date] = field(default=None)

    # Counts
    users_count: Optional[int] = field(default=None)
    fields_count: Optional[int] = field(default=None)
    resolutions_count: Optional[int] = field(default=None)
    issue_types_count: Optional[int] = field(default=None)
    issue_link_types_count: Optional[int] = field(default=None)
    priorities_count: Optional[int] = field(default=None)
    projects_count: Optional[int] = field(default=None)
    project_versions_count: Optional[int] = field(default=None)
    boards_count: Optional[int] = field(default=None)
    # Adheres to Pull From Date and (in remote manifests)
    # a project that we have excluded is zero'd out,
    # i.e. even if we can see issues in that project we
    # consider it to have 0 issues since we won't ingest any
    issues_count: Optional[int] = field(default=None)

    # Drill down into each project with ProjectManifests
    project_manifests: IJiraProjectManifestList = field(default_factory=list)

    # For debug purposes. We may want to optionally exclude this when serializing
    encountered_errors_for_projects: dict = field(default_factory=dict)


IJiraDataManifest = TypeVar('IJiraDataManifest', bound='JiraDataManifestBase')



from datetime import datetime
from unittest.mock import patch

import pytest
import pytz
import requests
import requests_mock

from jf_ingest.config import GithubAuthConfig
from jf_ingest.graphql_utils import (
    GQL_PAGE_INFO_BLOCK,
    GQLRateLimit,
    gql_format_to_datetime,
)
from jf_ingest.jf_git.clients.github import GQL_RATE_LIMIT_QUERY_BLOCK, GithubClient
from jf_ingest.jf_git.exceptions import GqlRateLimitExceededException

TEST_BASE_URL = 'https://gav.com/gql'

# Shared test data
GQL_DATETIME_STR = "2024-12-29T12:00:58Z"


@pytest.fixture()
def client() -> GithubClient:
    # Mock out a Github GQL Client
    github_auth_config = GithubAuthConfig(company_slug='a company', base_url=TEST_BASE_URL, token='test')
    return GithubClient(github_auth_config)


def test_gql_page_results_block():
    """This is a sanity check that the GQL_PAGE_INFO_BLOCK and GQL_RATE_LIMIT_QUERY_BLOCK
    are valid. This are pretty vague smoke tests... it's unlikely anybody would mess with
    these, but if they do it will cause a lot of problems!
    """

    assert 'pageInfo' in GQL_PAGE_INFO_BLOCK
    assert 'hasNextPage' in GQL_PAGE_INFO_BLOCK
    assert 'endCursor' in GQL_PAGE_INFO_BLOCK

    assert 'rateLimit' in GQL_RATE_LIMIT_QUERY_BLOCK
    assert 'remaining' in GQL_RATE_LIMIT_QUERY_BLOCK
    assert 'resetAt' in GQL_RATE_LIMIT_QUERY_BLOCK


def test_gql_format_to_datetime():
    """Assert we get valid date time back (and that it's timezone aware)"""
    gql_datetime: datetime = gql_format_to_datetime(datetime_str=GQL_DATETIME_STR)
    assert type(gql_datetime) is datetime
    assert gql_datetime.tzinfo is not None
    assert gql_datetime == datetime(2024, 12, 29, 12, 00, 58, tzinfo=pytz.utc)

def test_git_auth_config_base_url():
    github_api_url = 'https://api.github.com'
    github_auth_config = GithubAuthConfig(company_slug='a company', token='test')
    assert github_auth_config.base_url == GithubAuthConfig.base_url
    assert github_auth_config.base_url == None

    client = GithubClient(github_auth_config)
    assert client.rest_api_url == github_api_url
    assert client.gql_base_url == f'{github_api_url}/graphql'

def test_get_rate_limit(requests_mock: requests_mock.Mocker, client: GithubClient):
    # Mock classes/data
    session = requests.Session()
    client.session = session
    
    def _test_get_rate_limit_inner(remaining_rate_limit: int, reset_at_rate_limit_str: str):
        print(
            f'Testing get_rate_limit with remaining={remaining_rate_limit} and resetAt={reset_at_rate_limit_str}'
        )
        # Mock classes/data
        session_mock_return_data = {
            'data': {
                'rateLimit': {'remaining': remaining_rate_limit, 'resetAt': reset_at_rate_limit_str}
            }
        }
        requests_mock.post(url=f'{TEST_BASE_URL}/graphql', json=session_mock_return_data)

        # Call function
        rate_limit_info = client.get_gql_rate_limit()

        # Test results
        assert type(rate_limit_info) is GQLRateLimit
        assert rate_limit_info.reset_at == gql_format_to_datetime(datetime_str=GQL_DATETIME_STR)
        assert rate_limit_info.remaining == int(remaining_rate_limit)

    # Test data
    _test_get_rate_limit_inner(25, GQL_DATETIME_STR)
    _test_get_rate_limit_inner(-1, GQL_DATETIME_STR)
    _test_get_rate_limit_inner("124", GQL_DATETIME_STR)
    with pytest.raises(Exception):
        _test_get_rate_limit_inner('cat', GQL_DATETIME_STR)


def test_get_raw_gql_result_simple(requests_mock: requests_mock.Mocker, client: GithubClient):
    # Test Data
    test_query_body = "test_query_body"
    # Mock classes/data
    session = requests.Session()
    session_mock_caller_data = {'query': test_query_body}
    session_mock_return_data = {'data': {'test': {'remaining': 'test', 'resetAt': 'test'}}}
    requests_mock.post(url=f'{TEST_BASE_URL}/graphql', json=session_mock_return_data)
    client.session = session

    # call function
    returned_data = client.get_raw_result(query_body=test_query_body)

    assert returned_data == session_mock_return_data
    assert requests_mock.last_request.json() == session_mock_caller_data


def test_get_raw_gql_result_force_error_raise(mocker, requests_mock: requests_mock.Mocker, client: GithubClient):
    mocker.patch('time.sleep', return_value=None)
    # Mock classes/data
    session = requests.Session()
    client.session = session

    requests_mock.post(
        url=f'{TEST_BASE_URL}/graphql', response_list=[{'json': '', 'status_code': 403}]
    )

    print('Testing raising a 403 error intentionally')
    with pytest.raises(Exception):
        print('Max attempts is 0...')
        client.get_raw_result(query_body='', max_attempts=0)
        print('Max attempts is 100... (if time.sleep is NOT mocked here, this will hang forever!)')
        client.get_raw_result(query_body='', max_attempts=100)

    print('Testing raising a 500 error intentionally')
    requests_mock.post(url=TEST_BASE_URL, response_list=[{'json': '', 'status_code': 500}])
    with pytest.raises(Exception):
        client.get_raw_result(query_body='', max_attempts=0)
        client.get_raw_result(query_body='', max_attempts=100)

    print('Asserting that we hit the GQL Rate Limit Exceeded exception')
    reponse_to_force_retry = {'json': {'errors': [{'type': 'RATE_LIMITED'}]}, 'status_code': 200}
    response_for_retry_rate = {
        'json': {'data': {'rateLimit': {'remaining': 10, 'resetAt': GQL_DATETIME_STR}}},
        'status_code': 200,
    }
    reponse_pairing = [reponse_to_force_retry, response_for_retry_rate]
    response_list = [*reponse_pairing, *reponse_pairing, *reponse_pairing]

    requests_mock.post(
        url=f'{TEST_BASE_URL}/graphql',
        response_list=response_list,
    )

    with pytest.raises(GqlRateLimitExceededException):
        client.get_raw_result(query_body='', max_attempts=2)


def test_get_raw_gql_result_successful_retry(mocker, requests_mock: requests_mock.Mocker, client: GithubClient):
    mocker.patch('time.sleep', return_value=None)
    session = requests.Session()
    successful_data = {'success': True}
    requests_mock.post(
        url=f'{TEST_BASE_URL}/graphql',
        response_list=[
            {'json': {'errors': [{'type': 'RATE_LIMITED'}]}, 'status_code': 200},
            {
                'json': {'data': {'rateLimit': {'remaining': 10, 'resetAt': GQL_DATETIME_STR}}},
                'status_code': 200,
            },
            {'json': successful_data, 'status_code': 200},
        ],
    )

    client.session = session

    data = client.get_raw_result(query_body='', max_attempts=10)
    assert data == successful_data


def test_github_client_error_raise_403(mocker, requests_mock: requests_mock.Mocker, client: GithubClient):
    mocker.patch('time.sleep', return_value=None)
    session = requests.Session()
    successful_data = {'success': True}
    requests_mock.post(
        url=f'{TEST_BASE_URL}/graphql',
        response_list=[
            {'status_code': 403},
        ],
    )

    # Mock out a Github GQL Client
    client.session = session

    retry_count = 10
    with pytest.raises(requests.HTTPError):
        client.get_raw_result(query_body='', max_attempts=retry_count)
        assert requests_mock.call_count == retry_count

def test_github_client_error_resolves_after_retry(mocker, requests_mock: requests_mock.Mocker, client: GithubClient):
    mocker.patch('time.sleep', return_value=None)
    session = requests.Session()
    successful_data = {'mantra': 'never give UP!'}
    response_list = [
        # A Beautiful Error Medley
        {'status_code': 403},
        {'status_code': 429},
        {'status_code': 500},
        {'status_code': 502},
        {'status_code': 503},
        {'status_code': 504},
        {'json': successful_data}
    ]
    retry_count = len(response_list)
    requests_mock.post(
        url=f'{TEST_BASE_URL}/graphql',
        response_list=response_list,
    )

    # Mock out a Github GQL Client
    client.session = session

    assert client.get_raw_result(query_body='', max_attempts=retry_count) == successful_data
    assert requests_mock.call_count == retry_count

def test_page_results_gql(mocker, requests_mock: requests_mock.Mocker):
    session = requests.Session()
    # Mock out a Github GQL Client
    github_auth_config = GithubAuthConfig(company_slug='A Company', base_url=TEST_BASE_URL, session=session, token='test')
    client = GithubClient(github_auth_config)
    client.session = session

    def _test_page_results_gql_intter(items_per_page: int, number_of_pages: int):
        print(
            f'Testing page_results_gql(items_per_page={items_per_page}, number_of_pages={number_of_pages})'
        )
        query_body = f"""{{
            baseQuery {{
                queryToPage(first: {items_per_page}, after: %s) {{
                    pageInfo {{hasNextPage, endCursor}}
                    testField
                }}
            }}
        }}
        """
        path_to_page_info = 'data.baseQuery.queryToPage'

        requests_mock.post(
            url=f'{TEST_BASE_URL}/graphql',
            response_list=[
                {
                    'json': {
                        'data': {
                            'baseQuery': {
                                'queryToPage': {
                                    'pageInfo': {
                                        'hasNextPage': i != number_of_pages - 1,
                                        'endCursor': f'endCursor_{i}',
                                    },
                                    'nodes': [
                                        {
                                            'testField': f'pageNumber={i}, nodeNumber={x}',
                                        }
                                        for x in range(0, items_per_page)
                                    ],
                                }
                            }
                        }
                    },
                    'status_code': 200,
                }
                for i in range(0, number_of_pages)
            ],
        )

        pages = [
            page
            for page in client.page_results_gql(
                query_body=query_body,
                path_to_page_info=path_to_page_info,
            )
        ]

        assert len(pages) == number_of_pages

        total_items = [
            item for page in pages for item in page['data']['baseQuery']['queryToPage']['nodes']
        ]

        assert len(total_items) == items_per_page * number_of_pages

    _test_page_results_gql_intter(items_per_page=1, number_of_pages=10)
    _test_page_results_gql_intter(items_per_page=10, number_of_pages=100)



import json
import logging
from contextlib import contextmanager
from datetime import datetime, timedelta, timezone
from typing import Dict
from unittest.mock import patch

import pytest
import pytz
from requests import Request
import requests_mock
from jira import JIRAError

from jf_ingest.constants import Constants
from jf_ingest.jf_jira.downloaders import (
    IssueMetadata,
    _download_issue_page,
    _expand_changelog,
    _filter_changelogs,
    _get_all_project_issue_counts,
    _get_batch_size_with_jql_enhanced_search,
    _get_exact_count_via_pagination,
    _get_issue_count_for_jql,
    _get_issue_count_with_jql_enhanced_search,
    _post_raw_result_jql_enhanced,
    is_jql_enhanced_search_available,
    fetch_id_to_key_for_all_existing,
    generate_jql_for_batch_of_ids,
    generate_project_pull_from_bulk_jql,
    generate_project_pull_from_jql,
    get_fields_spec,
    get_issues_with_post_enhanced,
    get_jira_results_looped,
    get_jira_search_batch_size,
    pull_all_jira_issues_by_date,
    pull_jira_issues_by_jira_ids,
)
from jf_ingest.jf_jira.utils import JiraFieldIdentifier
from jf_ingest.utils import batch_iterable
from tests.jf_jira.utils import (
    _register_jira_uri,
    _register_jira_uri_with_file,
    get_jira_mock_connection,
    get_jira_mock_download_config,
)

logger = logging.getLogger(__name__)


def _mock_server_info_endpoints(mocker):
    """Helper function to mock both v2 and v3 serverInfo endpoints"""
    server_info_response = '{"baseUrl":"https://test-co.atlassian.net","version":"1001.0.0-SNAPSHOT","versionNumbers":[1001,0,0],"deploymentType":"Cloud"}'
    _register_jira_uri(mocker, "serverInfo", server_info_response)
    # Mock v3 serverInfo endpoint
    mocker.register_uri(
        'GET',
        'https://test-co.atlassian.net/rest/api/3/serverInfo',
        text=server_info_response
    )


def _generate_mock_address_for_issue_jql(
    m: requests_mock.Mocker,
    jql_query: str,
    issue_count: int,
    start_at: int,
    max_results: int,
    issues: list[dict],
):
    _issues = [issue for issue in issues[start_at : min(start_at + max_results, len(issues))]]
    jira_return_val = f'{{"expand":"names,schema","startAt":{start_at},"maxResults":{max_results},"total":{issue_count},"issues":{json.dumps(_issues)}}}'

    endpoint = (
        f"search?jql={jql_query}&startAt={start_at}&validateQuery=True&maxResults={max_results}"
    )
    _register_jira_uri(
        m,
        endpoint=endpoint,
        return_value=jira_return_val,
    )


def _mock_jira_issue_by_date_endpoints(
    m: requests_mock.Mocker,
    project_keys_to_issue_counts: dict[str, int],
    pull_from: datetime,
    batch_size: int,
    issues_updated_value: datetime = pytz.utc.localize(datetime.min),
    expand_fields: list[str] = ["*all"],
):
    def generate_issues(project_key, count):
        _fields = {}
        if "*all" in expand_fields:
            _fields["updated"] = issues_updated_value.strftime("%Y-%m-%dT%H:%M:%S.000-0000")
        else:
            if "updated" in expand_fields:
                _fields["updated"] = issues_updated_value.strftime("%Y-%m-%dT%H:%M:%S.000-0000")
        return [
            {
                "expand": "operations,versionedRepresentations,editmeta,changelog",
                "id": f"{i}",
                "self": "https://test-co.atlassian.net/rest/api/2/issue/63847",
                "key": f"{project_key}-{i}",
                "fields": _fields,
            }
            for i in range(count)
        ]

    for project_key, count in project_keys_to_issue_counts.items():
        issues = generate_issues(project_key=project_key, count=count)
        jql_query = generate_project_pull_from_jql(project_key=project_key, pull_from=pull_from)
        # Generate one call for getting hte 'first' page (for issue counts)
        _generate_mock_address_for_issue_jql(
            m=m,
            jql_query=jql_query,
            issue_count=count,
            start_at=0,
            issues=issues,
            max_results=1,
        )
        for start_at in range(0, count, batch_size):
            _generate_mock_address_for_issue_jql(
                m=m,
                jql_query=jql_query,
                issue_count=count,
                start_at=start_at,
                max_results=batch_size,
                issues=issues,
            )


def test_get_issue_count_for_jql():
    pull_from = datetime.min
    PROJECT_KEY = "PROJ"
    PROJECT_ISSUE_COUNT = 5123
    project_key_to_count = {PROJECT_KEY: PROJECT_ISSUE_COUNT}

    with requests_mock.Mocker() as mocker:
        _mock_jira_issue_by_date_endpoints(
            m=mocker,
            project_keys_to_issue_counts=project_key_to_count,
            pull_from=pull_from,
            batch_size=Constants.MAX_ISSUE_API_BATCH_SIZE,
        )
        count_for_jql = _get_issue_count_for_jql(
            get_jira_mock_connection(mocker),
            jql_query=generate_project_pull_from_jql(project_key=PROJECT_KEY, pull_from=pull_from),
        )
        assert count_for_jql == project_key_to_count[PROJECT_KEY]


def test_get_issue_count_for_jql_400_level_error_handling():
    """Assert that when we raise 400 level errors, we always return 0"""
    for status_code in range(400, 500):
        with patch(
            "jf_ingest.jf_jira.downloaders.retry_for_status",
            side_effect=JIRAError(status_code=status_code),
        ):
            logger.info(
                f"Attempting to test _get_issue_count_for_jql when a {status_code} error is thrown"
            )
            count_for_jql = _get_issue_count_for_jql(get_jira_mock_connection(), jql_query="")
            assert count_for_jql == 0


def test_get_issue_count_for_jql_500_level_error_handling():
    for status_code in range(500, 600):
        logger.info(f"Checking to see if we raise 500 level errors...")
        with patch(
            "jf_ingest.jf_jira.downloaders.retry_for_status",
            side_effect=JIRAError(status_code=status_code),
        ):
            with pytest.raises(JIRAError):
                _get_issue_count_for_jql(get_jira_mock_connection(), jql_query="")


def test_get_all_project_issue_counts():
    pull_from = datetime.min
    project_keys_to_counts = {"PROJ": 151, "COLLAGE": 512}

    with requests_mock.Mocker() as mocker:
        _mock_jira_issue_by_date_endpoints(
            m=mocker,
            project_keys_to_issue_counts=project_keys_to_counts,
            pull_from=pull_from,
            batch_size=Constants.MAX_ISSUE_API_BATCH_SIZE,
        )
        project_keys_to_pull_from = {
            project_key: pull_from for project_key in project_keys_to_counts.keys()
        }
        project_issue_counts = _get_all_project_issue_counts(
            get_jira_mock_connection(mocker),
            project_key_to_pull_from=project_keys_to_pull_from,
            num_parallel_threads=1,
            jql_filter=None,
        )

        assert project_issue_counts == project_keys_to_counts


def _mock_jira_issue_by_ids(
    m: requests_mock.Mocker(),
    issue_ids: list[str],
    batch_size: int,
    issues_updated_value: datetime = datetime.min,
    expand_fields: list[str] = ["*all"],
):
    def _generate_issues(ids_batch):
        _fields = {}
        if "*all" in expand_fields:
            _fields["updated"] = issues_updated_value.isoformat()
            _fields["parent"] = {"id": "PARENT", "key": f"PROJ-PARENT"}
        else:
            if "updated" in expand_fields:
                _fields["updated"] = issues_updated_value.isoformat()
            if "parent" in expand_fields:
                _fields["parent"] = {"id": "PARENT", "key": f"PROJ-PARENT"}

        return [
            {
                "expand": "operations,versionedRepresentations,editmeta,changelog",
                "id": f"{id}",
                "self": "https://test-co.atlassian.net/rest/api/2/issue/63847",
                "key": f"PROJ-{i}",
                "fields": _fields,
            }
            for i, id in enumerate(ids_batch)
        ]

    for id_batch in batch_iterable(sorted(issue_ids, key=int), batch_size=batch_size):
        jql_query = generate_jql_for_batch_of_ids(id_batch)
        _generate_mock_address_for_issue_jql(
            m=m,
            jql_query=jql_query,
            issue_count=len(id_batch),
            start_at=0,
            issues=_generate_issues(id_batch),
            max_results=batch_size,
        )

def test_generate_jql_for_batch_ids():
    ids = ["1", "2", "3", "4", "5"]
    jql_query = generate_jql_for_batch_of_ids(ids)
    assert jql_query == 'id in (1,2,3,4,5) order by id asc'

def test_generate_jql_for_batch_ids_with_pull_from():
    ids = ["1", "2", "3", "4", "5"]
    pull_from = datetime(2024, 12, 29, 17, 20, 1)
    jql_query = generate_jql_for_batch_of_ids(ids, pull_from)
    assert jql_query == 'id in (1,2,3,4,5) AND updated > "2024-12-29" order by id asc'


def test_get_jira_batch_size():
    @contextmanager
    def _mocked_jira_return(requested_batch_size: int, returned_batch_size: int):
        with requests_mock.Mocker() as mocker:
            jira_return_val = f'{{"expand":"names,schema","startAt":0,"maxResults":{returned_batch_size},"total":0,"issues":[]}}'

            _register_jira_uri(
                mocker,
                endpoint=f"search",
                return_value=jira_return_val,
                HTTP_ACTION='POST',
            )
            yield mocker

    optimistic_batch_size = 1000
    for jira_batch_size_return in [0, 10, Constants.MAX_ISSUE_API_BATCH_SIZE, 1000, 1235]:
        with _mocked_jira_return(
            requested_batch_size=optimistic_batch_size,
            returned_batch_size=jira_batch_size_return,
        ) as mocker:
            # Check when fields is left out (it should default to [*all])
            jira_issues_batch_size = get_jira_search_batch_size(
                jira_connection=get_jira_mock_connection(mocker),
                optimistic_batch_size=optimistic_batch_size,
            )

            assert jira_issues_batch_size == jira_batch_size_return


def test_get_jira_batch_size_with_variable_field_argument():
    @contextmanager
    def _mocked_jira_return(requested_batch_size: int, returned_batch_size: int):
        with requests_mock.Mocker() as mocker:
            jira_return_val = f'{{"expand":"names,schema","startAt":0,"maxResults":{returned_batch_size},"total":0,"issues":[]}}'

            _register_jira_uri(
                mocker,
                endpoint=f"search",
                return_value=jira_return_val,
                HTTP_ACTION='POST',
            )
            yield mocker

    optimistic_batch_size = 1000
    for jira_batch_size_return in [0, 10, Constants.MAX_ISSUE_API_BATCH_SIZE, 1000, 1235]:
        with _mocked_jira_return(
            requested_batch_size=optimistic_batch_size,
            returned_batch_size=jira_batch_size_return,
        ) as mocker:
            # Check when fields is left out (it should default to [*all])
            get_jira_search_batch_size(
                jira_connection=get_jira_mock_connection(mocker),
                optimistic_batch_size=optimistic_batch_size,
            )

            def _get_request_body():
                return json.loads(mocker.request_history[-1]._request.body)

            print(json.loads(mocker.request_history[-1]._request.body)['fields'])
            assert _get_request_body()['fields'] == ['*all']

            # Check when fields is set to ['id', 'key']
            fields = ['key', 'id']
            get_jira_search_batch_size(
                jira_connection=get_jira_mock_connection(mocker),
                optimistic_batch_size=optimistic_batch_size,
                fields=fields,
            )

            assert _get_request_body()['fields'] == fields

            # Check when fields is set to ['*all'] manually
            fields = ['*all']
            get_jira_search_batch_size(
                jira_connection=get_jira_mock_connection(mocker),
                optimistic_batch_size=optimistic_batch_size,
                fields=fields,
            )
            assert _get_request_body()['fields'] == fields


def test_get_jira_batch_size_with_jql_enhanced_search():
    """Test get_jira_search_batch_size with JQL Enhanced Search API."""
    optimistic_batch_size = 1000
    
    # Test ID-only fields with JQL Enhanced Search
    batch_size = get_jira_search_batch_size(
        jira_connection=None,  # Not used when use_jql_enhanced_search=True
        optimistic_batch_size=optimistic_batch_size,
        fields=['id'],
        use_jql_enhanced_search=True,
    )
    assert batch_size == 5000  # ID-only queries use higher limit
    
    # Test ID and key fields with JQL Enhanced Search
    batch_size = get_jira_search_batch_size(
        jira_connection=None,  # Not used when use_jql_enhanced_search=True
        optimistic_batch_size=optimistic_batch_size,
        fields=['id', 'key'],
        use_jql_enhanced_search=True,
    )
    assert batch_size == 5000  # ID-only queries use higher limit
    
    # Test full fields with JQL Enhanced Search
    batch_size = get_jira_search_batch_size(
        jira_connection=None,  # Not used when use_jql_enhanced_search=True
        optimistic_batch_size=optimistic_batch_size,
        fields=['*all'],
        use_jql_enhanced_search=True,
    )
    assert batch_size == 100  # Full issue queries use lower limit
    
    # Test with optimistic batch size smaller than default
    batch_size = get_jira_search_batch_size(
        jira_connection=None,  # Not used when use_jql_enhanced_search=True
        optimistic_batch_size=50,
        fields=['*all'],
        use_jql_enhanced_search=True,
    )
    assert batch_size == 50  # Should use the smaller optimistic batch size


def test_get_jira_batch_size_legacy_api_unchanged():
    """Test that legacy API behavior is unchanged when use_jql_enhanced_search=False."""
    @contextmanager
    def _mocked_jira_return(requested_batch_size: int, returned_batch_size: int):
        with requests_mock.Mocker() as mocker:
            jira_return_val = f'{{"expand":"names,schema","startAt":0,"maxResults":{returned_batch_size},"total":0,"issues":[]}}'

            _register_jira_uri(
                mocker,
                endpoint=f"search",
                return_value=jira_return_val,
                HTTP_ACTION='POST',
            )
            yield mocker

    optimistic_batch_size = 1000
    for jira_batch_size_return in [0, 10, Constants.MAX_ISSUE_API_BATCH_SIZE, 1000, 1235]:
        with _mocked_jira_return(
            requested_batch_size=optimistic_batch_size,
            returned_batch_size=jira_batch_size_return,
        ) as mocker:
            # Test with use_jql_enhanced_search=False (default)
            jira_issues_batch_size = get_jira_search_batch_size(
                jira_connection=get_jira_mock_connection(mocker),
                optimistic_batch_size=optimistic_batch_size,
                use_jql_enhanced_search=False,
            )
            assert jira_issues_batch_size == jira_batch_size_return
            
            # Test with default parameter (should be False)
            jira_issues_batch_size_default = get_jira_search_batch_size(
                jira_connection=get_jira_mock_connection(mocker),
                optimistic_batch_size=optimistic_batch_size,
            )
            assert jira_issues_batch_size_default == jira_batch_size_return


def test_get_fields_spec():
    assert get_fields_spec(include_fields=[], exclude_fields=[]) == ["*all"]
    assert get_fields_spec(include_fields=["updated"], exclude_fields=[]) == ["updated"]
    assert get_fields_spec(include_fields=["updated", "parent"], exclude_fields=[]) == [
        "updated",
        "parent",
    ]
    assert get_fields_spec(include_fields=["updated"], exclude_fields=["parent"]) == [
        "updated",
        "-parent",
    ]


def get_issues_through_test_fixture():
    issue_ids = sorted(["18447", "18404", "18031", "18018", "18016"], key=int)
    jql_query = generate_jql_for_batch_of_ids(issue_ids)
    with requests_mock.Mocker() as m:
        # Register one endpoint that this will hit
        uri = f"search?jql={jql_query}&startAt=0&validateQuery=True&fields=%2Aall&expand=changelog&maxResults=5"
        _register_jira_uri_with_file(m, endpoint=uri, fixture_path="api_responses/issues.json")

        return [
            i
            for i in pull_jira_issues_by_jira_ids(
                jira_connection=get_jira_mock_connection(),
                jira_ids=issue_ids,
                num_parallel_threads=10,
                batch_size=len(issue_ids),
                expand_fields=["changelog"],
                include_fields=[],
                exclude_fields=[],
            )
        ]

CUSTOM_FIELDS_FOR_FILTERING = tuple(["customfield_10051", "customfield_10057", "customfield_10009"])
JFI_FIELDS_FOR_FILTERING = tuple([
    JiraFieldIdentifier(jira_field_id=jira_id, jira_field_name=f'Name: {jira_id}')
    for jira_id in CUSTOM_FIELDS_FOR_FILTERING
])
def test_filter_changelogs_no_filtering():
    issues = get_issues_through_test_fixture()
    issues_without_filtering = _filter_changelogs(issues, [], [])

    for issue in issues_without_filtering:
        for history in issue["changelog"]["histories"]:
            assert len(history["items"]) != 0
            for item in history["items"]:
                if "fieldId" in item:
                    assert item["fieldId"] in CUSTOM_FIELDS_FOR_FILTERING


def test_filter_changelogs_inclusion_filtering_for_madeup_field():
    issues = get_issues_through_test_fixture()
    issues_with_filtering_field_in = _filter_changelogs(issues, [JiraFieldIdentifier(jira_field_id="madeup_field", jira_field_name='Made up Field')], [])
    assert len(issues) == len(issues_with_filtering_field_in)
    for issue in issues_with_filtering_field_in:
        for history in issue["changelog"]["histories"]:
            assert len(history["items"]) == 0

def test_filter_changelogs_exclusion_filtering_for_madeup_field():
    issues = get_issues_through_test_fixture()
    issues_with_filtering_field_in = _filter_changelogs(issues, [], [JiraFieldIdentifier(jira_field_id="madeup_field", jira_field_name='Made up Field')])
    assert len(issues) == len(issues_with_filtering_field_in)
    for issue, filtered_issue in zip(issues, issues_with_filtering_field_in):
        assert len(issue['changelog']['histories']) == len(filtered_issue['changelog']['histories'])
        for history, filtered_history in zip(issue["changelog"]["histories"], filtered_issue['changelog']['histories']):
            assert len(history["items"]) == len(filtered_history['items'])

def test_filter_changelogs_inclusion_filtering_by_id():
    field_id_1 = 'FIELD_ID_1'
    field_name_1 = 'FIELD NAME 1'
    field_id_2 = 'FIELD_ID_2'
    field_name_2 = 'FIELD NAME 2'
    issue = {
            'changelog': {
                'histories': [
                        {
                        'items': [
                            {
                                'fieldId': field_id_1,
                                'field': field_name_1,
                            },
                            {
                                'fieldId': field_id_2,
                                'field': field_name_2,
                            }
                        ]
                    }
                ]
            }
        }
    include_fields = [
        JiraFieldIdentifier(jira_field_id=field_id_1, jira_field_name=field_name_1)
    ]
    filtered_issue = _filter_changelogs([issue], include_fields, [])[0]
    print(filtered_issue)
    for history in filtered_issue["changelog"]["histories"]:
        assert len(history["items"]) == 1
        assert history['items'][0]['fieldId'] == field_id_1

def test_filter_changelogs_inclusion_filtering_by_name():
    field_id_1 = 'FIELD_ID_1'
    field_name_1 = 'FIELD NAME 1'
    field_id_2 = 'FIELD_ID_2'
    field_name_2 = 'FIELD NAME 2'
    issue = {
            'changelog': {
                'histories': [
                        {
                        'items': [
                            {
                                'field': field_name_1,
                            },
                            {
                                'fieldId': field_id_2,
                                'field': field_name_2,
                            }
                        ]
                    }
                ]
            }
        }
    include_fields = [
        JiraFieldIdentifier(jira_field_id=field_id_1, jira_field_name=field_name_1)
    ]
    filtered_issue = _filter_changelogs([issue], include_fields, [])[0]
    print(filtered_issue)
    for history in filtered_issue["changelog"]["histories"]:
        assert len(history["items"]) == 1
        assert history['items'][0]['field'] == field_name_1

def test_filter_changelogs_exclusion_filtering_by_id():
    field_id_1 = 'FIELD_ID_1'
    field_name_1 = 'FIELD NAME 1'
    field_id_2 = 'FIELD_ID_2'
    field_name_2 = 'FIELD NAME 2'
    issue = {
            'changelog': {
                'histories': [
                        {
                        'items': [
                            {
                                'field': field_name_1,
                                'fieldId': field_id_1
                            },
                            {
                                'fieldId': field_id_2,
                                'field': field_name_2,
                            }
                        ]
                    }
                ]
            }
        }
    exclude_fields = [
        JiraFieldIdentifier(jira_field_id=field_id_2, jira_field_name=field_name_2)
    ]
    filtered_issue = _filter_changelogs([issue], [], exclude_fields)[0]
    print(filtered_issue)
    for history in filtered_issue["changelog"]["histories"]:
        assert len(history["items"]) == 1
        assert history['items'][0]['fieldId'] == field_id_1

def test_filter_changelogs_exclusion_filtering_by_name():
    field_id_1 = 'FIELD_ID_1'
    field_name_1 = 'FIELD NAME 1'
    field_id_2 = 'FIELD_ID_2'
    field_name_2 = 'FIELD NAME 2'
    issue = {
            'changelog': {
                'histories': [
                        {
                        'items': [
                            {
                                'field': field_name_1,
                            },
                            {
                                'fieldId': field_id_2,
                                'field': field_name_2,
                            }
                        ]
                    }
                ]
            }
        }
    exclude_fields = [
        JiraFieldIdentifier(jira_field_id=field_id_2, jira_field_name=field_name_2)
    ]
    filtered_issue = _filter_changelogs([issue], [], exclude_fields)[0]
    print(filtered_issue)
    for history in filtered_issue["changelog"]["histories"]:
        assert len(history["items"]) == 1
        assert history['items'][0]['field'] == field_name_1

@pytest.mark.skip(reason="need to mock serverInfo endpoint too")
def test_expand_changelog():
    total_changelog_histories = 5
    batch_size = 1

    def _mock_api_endpoint_for_changelog(m: requests_mock.Mocker, change_log_num: int):
        mock_return = {
            "self": "https://test-co.atlassian.net/rest/api/2/issue/TS-4/changelog?maxResults=1&startAt=1",
            "nextPage": "https://test-co.atlassian.net/rest/api/2/issue/TS-4/changelog?maxResults=1&startAt=2",
            "maxResults": batch_size,
            "startAt": change_log_num - 1,
            "total": total_changelog_histories,
            "isLast": False,
            "values": [
                {
                    "id": f"{change_log_num}",
                    "author": {},
                    "created": "2020-06-29T16:01:51.141-0400",
                    "items": [
                        {
                            "field": "Spelunking CustomField v2",
                            "fieldtype": "custom",
                            "fieldId": "customfield_10057",
                            "from": None,
                            "fromString": None,
                            "to": "10072",
                            "toString": "hello",
                        }
                    ],
                }
            ],
        }
        _register_jira_uri(
            m,
            endpoint=f"issue/1/changelog?startAt={change_log_num - 1}&maxResults={batch_size}",
            return_value=json.dumps(mock_return),
        )

    with requests_mock.Mocker() as m:
        for change_log_num in range(0, total_changelog_histories + 1):
            _mock_api_endpoint_for_changelog(m, change_log_num)
        spoofed_issue_raw: dict = {
            "id": "1",
            "key": "spoof-1",
            "changelog": {
                "total": total_changelog_histories,
                "maxResults": 0,
                "histories": [],
            },
        }

        spoofed_issue_no_more_results_raw: dict = {
            "id": "2",
            "key": "spoof-2",
            "changelog": {"total": 0, "maxResults": 0, "histories": []},
        }

        _expand_changelog(
            get_jira_mock_connection(),
            jira_issues=[spoofed_issue_raw, spoofed_issue_no_more_results_raw],
            batch_size=1,
        )

        assert len(spoofed_issue_raw["changelog"]["histories"]) == total_changelog_histories
        assert len(spoofed_issue_no_more_results_raw["changelog"]["histories"]) == 0


@contextmanager
def _mock_for_full_issue_test(
    jf_issue_metadata: list[IssueMetadata],
    project_key: str = "PROJ",
    pull_from: datetime = datetime.min,
    issues_updated_value: datetime = datetime(2020, 1, 1),
    batch_size: int = Constants.MAX_ISSUE_API_BATCH_SIZE,
):
    expand_fields = ["*all"]

    with requests_mock.Mocker() as mocker:
        # Register the 'Batch Size' query return
        _register_jira_uri(
            mocker,
            endpoint=f"search?jql=&startAt=0&validateQuery=True&fields=%2Aall&maxResults={Constants.MAX_ISSUE_API_BATCH_SIZE}",
            return_value=f'{{"expand":"schema,names","startAt":0,"maxResults":{batch_size},"total":{len(jf_issue_metadata)},"issues":[]}}',
        )

        # Register the 'pull from' dates
        _mock_jira_issue_by_date_endpoints(
            m=mocker,
            project_keys_to_issue_counts={project_key: len(jf_issue_metadata)},
            pull_from=pull_from,
            issues_updated_value=issues_updated_value,
            batch_size=Constants.MAX_ISSUE_API_BATCH_SIZE,
        )

        _mock_jira_issue_by_ids(
            m=mocker,
            issue_ids=[
                issue_metadata.id
                for issue_metadata in jf_issue_metadata
                if issues_updated_value > issue_metadata.updated
            ],
            batch_size=batch_size,
            issues_updated_value=issues_updated_value,
            expand_fields=expand_fields,
        )
        yield mocker


def test_download_issue_page_ensure_error_never_raised():
    """
    The _download_issue_page should NEVER raise an error.
    """
    with patch(
        "jf_ingest.jf_jira.downloaders.retry_for_status",
        side_effect=JIRAError(status_code=500),
    ):
        issues = _download_issue_page(
            jira_connection=get_jira_mock_connection(), jql_query='', batch_size=100, start_at=0
        )
        assert len(issues) == 0

    with patch(
        "jf_ingest.jf_jira.downloaders.retry_for_status",
        side_effect=Exception('random exception'),
    ):
        issues = _download_issue_page(
            jira_connection=get_jira_mock_connection(), jql_query='', batch_size=100, start_at=0
        )
        assert len(issues) == 0

def test_download_issue_page_ensure_fields_not_ignored():
    """
    The _download_issue_page should NEVER raise an error.
    """
    @contextmanager
    def _mocked_jira_return(requested_batch_size: int, returned_batch_size: int):
        with requests_mock.Mocker() as mocker:
            jira_return_val = f'{{"expand":"names,schema","startAt":0,"maxResults":0,"total":0,"issues":[]}}'
            jira_mock_connection = get_jira_mock_connection(mocker)
            _register_jira_uri(
                mocker,
                endpoint=f"search",
                return_value=jira_return_val,
                HTTP_ACTION='POST',
            )
            yield jira_mock_connection, mocker

    with _mocked_jira_return(requested_batch_size=0, returned_batch_size=0) as (jira_conn, mocker):
        _download_issue_page(
            jira_connection=jira_conn, jql_query='', batch_size=0, start_at=0, expand_fields=[], include_fields=[
                JiraFieldIdentifier(jira_field_id='id', jira_field_name='ID')
            ]
        )
        print(vars(mocker.request_history[0]))
        post_found = False
        for request in mocker.request_history:
            request: Request = request._request
            if request.method == 'POST':
                post_found = True
                request_body = json.loads(request.body)
                assert request_body['fields'] == ['id']
        assert post_found

def test_download_issue_page_ensure_fields_not_ignored_more():
    """
    The _download_issue_page should NEVER raise an error.
    """
    @contextmanager
    def _mocked_jira_return(requested_batch_size: int, returned_batch_size: int):
        with requests_mock.Mocker() as mocker:
            jira_return_val = f'{{"expand":"names,schema","startAt":0,"maxResults":0,"total":0,"issues":[]}}'
            jira_mock_connection = get_jira_mock_connection(mocker)
            _register_jira_uri(
                mocker,
                endpoint=f"search",
                return_value=jira_return_val,
                HTTP_ACTION='POST',
            )
            yield jira_mock_connection, mocker

    with _mocked_jira_return(requested_batch_size=0, returned_batch_size=0) as (jira_conn, mocker):
        _download_issue_page(
            jira_connection=jira_conn, jql_query='', batch_size=0, start_at=0, expand_fields=[], include_fields=[
                JiraFieldIdentifier(jira_field_id='id', jira_field_name='ID'),
                JiraFieldIdentifier(jira_field_id='customfield_101234', jira_field_name='Sprint'),
                JiraFieldIdentifier(jira_field_id='customfield_201244', jira_field_name='Thing'),
            ]
        )
        print(vars(mocker.request_history[0]))
        post_found = False
        for request in mocker.request_history:
            request: Request = request._request
            if request.method == 'POST':
                post_found = True
                request_body = json.loads(request.body)
                assert request_body['fields'] == ['id', 'customfield_101234', 'customfield_201244']
        assert post_found

def test_generate_project_pull_from_bulk_jql_base():
    project_keys = ['A', 'B', 'C', 'D', 'E', 'F']
    project_key_to_pull_from = {
        'A': datetime(2024, 10, 1),
        'B': datetime(2024, 10, 2),
        'C': datetime(2024, 10, 3),
        'D': datetime(2024, 10, 4),
        'E': datetime(2024, 10, 5),
        'F': datetime(2024, 10, 6),
    }
    jql_filters = [
        generate_project_pull_from_bulk_jql(project_keys=project_key_batch, project_key_to_pull_from=project_key_to_pull_from)
        for project_key_batch in batch_iterable(project_keys, batch_size=3)
    ]
    
    assert len(jql_filters) == 2
    assert jql_filters[0] == '(project = A AND updated > "2024-10-01") OR (project = B AND updated > "2024-10-02") OR (project = C AND updated > "2024-10-03") order by id asc'
    assert jql_filters[1] == '(project = D AND updated > "2024-10-04") OR (project = E AND updated > "2024-10-05") OR (project = F AND updated > "2024-10-06") order by id asc'

def test_generate_project_pull_from_bulk_jql_with_issue_filter():
    project_keys = ['A', 'B', 'C', 'D', 'E', 'F']
    project_key_to_pull_from = {
        'A': datetime(2024, 10, 1),
        'B': datetime(2024, 10, 2),
        'C': datetime(2024, 10, 3),
        'D': datetime(2024, 10, 4),
        'E': datetime(2024, 10, 5),
        'F': datetime(2024, 10, 6),
    }
    jql_filter = generate_project_pull_from_bulk_jql(project_keys=project_keys, project_key_to_pull_from=project_key_to_pull_from, jql_filter='issuetype != "Secret Type"')
    assert jql_filter == (
        '(project = A AND updated > "2024-10-01") OR (project = B AND updated > "2024-10-02") OR (project = C AND updated > "2024-10-03") OR (project = D AND updated > "2024-10-04") OR (project = E AND updated > "2024-10-05") OR (project = F AND updated > "2024-10-06") AND (issuetype != "Secret Type") order by id asc'
    )


def test_is_jql_enhanced_search_available_success():
    """Test that is_jql_enhanced_search_available returns True when /search/jql endpoint returns 200"""
    with requests_mock.Mocker() as mocker:
        # Mock successful response from /search/jql endpoint (v3 API)
        jira_return_val = '{"maxResults": 1, "nextPageToken": "eyJzdGFydEF0IjoxfQ==", "issues": []}'
        mocker.register_uri(
            'POST',
            'https://test-co.atlassian.net/rest/api/3/search/jql',
            text=jira_return_val,
            status_code=200,
        )
        
        # Mock serverInfo endpoints for connection creation
        _mock_server_info_endpoints(mocker)
        
        jira_config = get_jira_mock_download_config()
        result = is_jql_enhanced_search_available(
            jira_config=jira_config,
            jql_enhanced_search_enabled=True,
            force_legacy_api=False
        )
        assert result is True


def test_is_jql_enhanced_search_available_not_found():
    """Test that is_jql_enhanced_search_available returns False when /search/jql endpoint returns 404"""
    with requests_mock.Mocker() as mocker:
        # Mock 404 response from /search/jql endpoint (v3 API)
        mocker.register_uri(
            'POST',
            'https://test-co.atlassian.net/rest/api/3/search/jql',
            text='{"errorMessages":["Not Found"]}',
            status_code=404,
        )
        
        # Mock serverInfo endpoints for connection creation
        _mock_server_info_endpoints(mocker)
        
        jira_config = get_jira_mock_download_config()
        result = is_jql_enhanced_search_available(
            jira_config=jira_config,
            jql_enhanced_search_enabled=True,
            force_legacy_api=False
        )
        assert result is False


def test_is_jql_enhanced_search_available_server_error():
    """Test that is_jql_enhanced_search_available returns False when /search/jql endpoint returns 500 (conservative fallback)"""
    with requests_mock.Mocker() as mocker:
        # Mock 500 response from /search/jql endpoint (v3 API)
        mocker.register_uri(
            'POST',
            'https://test-co.atlassian.net/rest/api/3/search/jql',
            text='{"errorMessages":["Internal Server Error"]}',
            status_code=500,
        )
        
        # Mock serverInfo endpoints for connection creation
        _mock_server_info_endpoints(mocker)
        
        jira_config = get_jira_mock_download_config()
        result = is_jql_enhanced_search_available(
            jira_config=jira_config,
            jql_enhanced_search_enabled=True,
            force_legacy_api=False
        )
        assert result is False


def test_is_jql_enhanced_search_available_auth_error():
    """Test that is_jql_enhanced_search_available returns False when /search/jql endpoint returns 401 (conservative fallback)"""
    with requests_mock.Mocker() as mocker:
        # Mock 401 response from /search/jql endpoint (v3 API)
        mocker.register_uri(
            'POST',
            'https://test-co.atlassian.net/rest/api/3/search/jql',
            text='{"errorMessages":["Unauthorized"]}',
            status_code=401,
        )
        
        # Mock serverInfo endpoints for connection creation
        _mock_server_info_endpoints(mocker)
        
        jira_config = get_jira_mock_download_config()
        result = is_jql_enhanced_search_available(
            jira_config=jira_config,
            jql_enhanced_search_enabled=True,
            force_legacy_api=False
        )
        assert result is False


def test_is_jql_enhanced_search_available_network_exception():
    """Test that is_jql_enhanced_search_available returns False when network exception occurs"""
    import requests
    from unittest.mock import MagicMock
    
    # Mock get_jira_connection to avoid connection creation delays
    with patch('jf_ingest.jf_jira.downloaders.get_jira_connection') as mock_get_connection:
        # Create a mock connection that raises an exception when _session.post is called
        mock_connection = MagicMock()
        mock_connection._session.post.side_effect = requests.exceptions.ConnectionError("Connection failed")
        mock_connection._get_url.return_value = "https://test-co.atlassian.net/rest/api/3/search/jql"
        mock_get_connection.return_value = mock_connection
        
        jira_config = get_jira_mock_download_config()
        result = is_jql_enhanced_search_available(
            jira_config=jira_config,
            jql_enhanced_search_enabled=True,
            force_legacy_api=False
        )
        assert result is False


def test_is_jql_enhanced_search_available_feature_flag_disabled():
    """Test that is_jql_enhanced_search_available returns False when jql_enhanced_search_enabled is False"""
    jira_config = get_jira_mock_download_config()
    result = is_jql_enhanced_search_available(
        jira_config=jira_config,
        jql_enhanced_search_enabled=False,
        force_legacy_api=False
    )
    assert result is False


def test_is_jql_enhanced_search_available_force_legacy_api():
    """Test that is_jql_enhanced_search_available returns False when force_legacy_api is True"""
    jira_config = get_jira_mock_download_config()
    result = is_jql_enhanced_search_available(
        jira_config=jira_config,
        jql_enhanced_search_enabled=True,
        force_legacy_api=True
    )
    assert result is False


def test_is_jql_enhanced_search_available_both_flags_false():
    """Test that is_jql_enhanced_search_available returns False when both flags are False"""
    jira_config = get_jira_mock_download_config()
    result = is_jql_enhanced_search_available(
        jira_config=jira_config,
        jql_enhanced_search_enabled=False,
        force_legacy_api=False
    )
    assert result is False


def test_post_raw_result_jql_enhanced_success():
    """Test that _post_raw_result_jql_enhanced successfully calls the /search/jql endpoint"""
    with requests_mock.Mocker() as mocker:
        # Mock successful response from /search/jql endpoint
        mock_response = {
            "maxResults": 5000,
            "nextPageToken": "eyJzdGFydEF0IjoxMDAwfQ==",
            "issues": [{"id": "1", "key": "TEST-1"}, {"id": "2", "key": "TEST-2"}]
        }
        _register_jira_uri(
            mocker,
            endpoint="search/jql",
            return_value=json.dumps(mock_response),
            HTTP_ACTION='POST',
            status_code=200,
        )
        
        jira_connection = get_jira_mock_connection(mocker)
        result = _post_raw_result_jql_enhanced(
            jira_connection=jira_connection,
            jql_query="project = TEST",
            fields=["id", "key"],
            expand=[],
            max_results=5000,
            next_page_token=None
        )
        
        assert result == mock_response
        assert result["maxResults"] == 5000
        assert result["nextPageToken"] == "eyJzdGFydEF0IjoxMDAwfQ=="
        assert len(result["issues"]) == 2


def test_post_raw_result_jql_enhanced_with_next_page_token():
    """Test that _post_raw_result_jql_enhanced correctly includes nextPageToken in request"""
    with requests_mock.Mocker() as mocker:
        # Mock successful response from /search/jql endpoint
        mock_response = {
            "maxResults": 5000,
            "issues": [{"id": "3", "key": "TEST-3"}]
        }
        
        def request_callback(request, context):
            # Verify that nextPageToken is included in the request
            request_data = json.loads(request.text)
            assert request_data["nextPageToken"] == "eyJzdGFydEF0IjoxMDAwfQ=="
            assert request_data["jql"] == "project = TEST"
            assert request_data["maxResults"] == 2500
            context.status_code = 200
            return json.dumps(mock_response)
        
        mocker.register_uri(
            'POST',
            'https://test-co.atlassian.net/rest/api/2/search/jql',
            text=request_callback
        )
        
        jira_connection = get_jira_mock_connection(mocker)
        result = _post_raw_result_jql_enhanced(
            jira_connection=jira_connection,
            jql_query="project = TEST",
            fields=["id"],
            expand=[],
            max_results=2500,
            next_page_token="eyJzdGFydEF0IjoxMDAwfQ=="
        )
        
        assert result == mock_response


def test_post_raw_result_jql_enhanced_retry_on_error():
    """Test that _post_raw_result_jql_enhanced retries on server errors"""
    with requests_mock.Mocker() as mocker:
        # First call returns 500, second call succeeds
        mock_response = {"maxResults": 5000, "issues": []}
        
        call_count = 0
        def request_callback(request, context):
            nonlocal call_count
            call_count += 1
            if call_count == 1:
                context.status_code = 500
                return '{"errorMessages":["Internal Server Error"]}'
            else:
                context.status_code = 200
                return json.dumps(mock_response)
        
        mocker.register_uri(
            'POST',
            'https://test-co.atlassian.net/rest/api/2/search/jql',
            text=request_callback
        )
        
        jira_connection = get_jira_mock_connection(mocker)
        result = _post_raw_result_jql_enhanced(
            jira_connection=jira_connection,
            jql_query="project = TEST",
            fields=["id"],
            expand=[],
            max_results=5000,
            next_page_token=None
        )
        
        assert result == mock_response
        assert call_count == 2  # Verify retry occurred


def test_get_issue_count_with_jql_enhanced_search_approximate_success():
    """Test that _get_issue_count_with_jql_enhanced_search uses approximate count when available"""
    with requests_mock.Mocker() as mocker:
        # Mock successful approximate count response (now uses POST)
        _register_jira_uri(
            mocker,
            endpoint="search/approximate-count",
            return_value='{"count": 1247}',
            HTTP_ACTION='POST',
            status_code=200,
        )
        
        jira_connection = get_jira_mock_connection(mocker)
        result = _get_issue_count_with_jql_enhanced_search(jira_connection, "project = TEST")
        
        assert result == 1247


def test_get_issue_count_with_jql_enhanced_search_fallback_to_pagination():
    """Test that _get_issue_count_with_jql_enhanced_search falls back to pagination when approximate count fails"""
    with requests_mock.Mocker() as mocker:
        # Mock failed approximate count response
        _register_jira_uri(
            mocker,
            endpoint="search/approximate-count",
            return_value='{"errorMessages":["Not supported"]}',
            HTTP_ACTION='GET',
            status_code=404,
        )
        
        # Mock pagination responses for exact counting
        # First page: 5000 issues with nextPageToken
        first_page_response = {
            "maxResults": 5000,
            "nextPageToken": "eyJzdGFydEF0Ijo1MDAwfQ==",
            "issues": [{"id": str(i)} for i in range(1, 5001)]  # 5000 issues
        }
        
        # Second page: 2500 issues, no nextPageToken (end of results)
        second_page_response = {
            "maxResults": 5000,
            "issues": [{"id": str(i)} for i in range(5001, 7501)]  # 2500 issues
        }
        
        call_count = 0
        def jql_request_callback(request, context):
            nonlocal call_count
            call_count += 1
            request_data = json.loads(request.text)
            
            if call_count == 1:
                # First call should have no nextPageToken
                assert "nextPageToken" not in request_data
                context.status_code = 200
                return json.dumps(first_page_response)
            else:
                # Second call should have nextPageToken
                assert request_data["nextPageToken"] == "eyJzdGFydEF0Ijo1MDAwfQ=="
                context.status_code = 200
                return json.dumps(second_page_response)
        
        mocker.register_uri(
            'POST',
            'https://test-co.atlassian.net/rest/api/2/search/jql',
            text=jql_request_callback
        )
        
        jira_connection = get_jira_mock_connection(mocker)
        result = _get_issue_count_with_jql_enhanced_search(jira_connection, "project = TEST")
        
        assert result == 7500  # 5000 + 2500
        assert call_count == 2  # Verify both pagination calls occurred


def test_get_issue_count_with_jql_enhanced_search_single_page():
    """Test that _get_issue_count_with_jql_enhanced_search handles single page results correctly"""
    with requests_mock.Mocker() as mocker:
        # Mock failed approximate count response
        _register_jira_uri(
            mocker,
            endpoint="search/approximate-count",
            return_value='{"errorMessages":["Not supported"]}',
            HTTP_ACTION='GET',
            status_code=404,
        )
        
        # Mock single page response (no nextPageToken)
        single_page_response = {
            "maxResults": 5000,
            "issues": [{"id": str(i)} for i in range(1, 101)]  # 100 issues
        }
        
        _register_jira_uri(
            mocker,
            endpoint="search/jql",
            return_value=json.dumps(single_page_response),
            HTTP_ACTION='POST',
            status_code=200,
        )
        
        jira_connection = get_jira_mock_connection(mocker)
        result = _get_issue_count_with_jql_enhanced_search(jira_connection, "project = TEST")
        
        assert result == 100


def test_get_batch_size_with_jql_enhanced_search_id_only():
    """Test that _get_batch_size_with_jql_enhanced_search returns 5000 for ID-only queries"""
    result = _get_batch_size_with_jql_enhanced_search(['id'], 10000)
    assert result == 5000
    
    result = _get_batch_size_with_jql_enhanced_search(['id', 'key'], 10000)
    assert result == 5000


def test_get_batch_size_with_jql_enhanced_search_full_issue_queries():
    """Test that _get_batch_size_with_jql_enhanced_search returns min(optimistic_batch_size, 100) for full issue queries"""
    # Test with many fields - should return min(optimistic_batch_size, 100)
    many_fields = ['id', 'key', 'summary', 'status', 'assignee', 'reporter']
    result = _get_batch_size_with_jql_enhanced_search(many_fields, 10000)
    assert result == 100
    
    # Test with few fields - should still return min(optimistic_batch_size, 100)
    few_fields = ['summary', 'status']
    result = _get_batch_size_with_jql_enhanced_search(few_fields, 10000)
    assert result == 100


def test_get_batch_size_with_jql_enhanced_search_respects_optimistic_limit():
    """Test that _get_batch_size_with_jql_enhanced_search behavior matches design specification"""
    # For ID queries, design shows hard-coded 5000 (doesn't respect optimistic_batch_size)
    result = _get_batch_size_with_jql_enhanced_search(['id'], 1000)
    assert result == 5000  # Design shows hard-coded 5000 for ID queries
    
    # For full issue queries, should return min(optimistic_batch_size, 100)
    result = _get_batch_size_with_jql_enhanced_search(['summary', 'status'], 50)
    assert result == 50  # Should respect the lower optimistic limit


def test_get_batch_size_with_jql_enhanced_search_with_iterable():
    """Test that _get_batch_size_with_jql_enhanced_search works with different iterable types"""
    # Test with tuple for ID queries
    result = _get_batch_size_with_jql_enhanced_search(('id', 'key'), 10000)
    assert result == 5000
    
    # Test with set for ID queries
    result = _get_batch_size_with_jql_enhanced_search({'id'}, 10000)
    assert result == 5000
    
    # Test with tuple for full issue queries
    result = _get_batch_size_with_jql_enhanced_search(('summary', 'status'), 10000)
    assert result == 100


def test_get_exact_count_via_pagination_single_page():
    """Test that _get_exact_count_via_pagination handles single page results correctly"""
    with requests_mock.Mocker() as mocker:
        # Mock single page response (no nextPageToken)
        single_page_response = {
            "maxResults": 5000,
            "issues": [{"id": str(i)} for i in range(1, 101)]  # 100 issues
        }
        
        _register_jira_uri(
            mocker,
            endpoint="search/jql",
            return_value=json.dumps(single_page_response),
            HTTP_ACTION='POST',
            status_code=200,
        )
        
        jira_connection = get_jira_mock_connection(mocker)
        result = _get_exact_count_via_pagination(jira_connection, "project = TEST")
        
        assert result == 100


def test_get_exact_count_via_pagination_multiple_pages():
    """Test that _get_exact_count_via_pagination handles multiple pages correctly"""
    with requests_mock.Mocker() as mocker:
        # Mock pagination responses
        # First page: 5000 issues with nextPageToken
        first_page_response = {
            "maxResults": 5000,
            "nextPageToken": "eyJzdGFydEF0Ijo1MDAwfQ==",
            "issues": [{"id": str(i)} for i in range(1, 5001)]  # 5000 issues
        }
        
        # Second page: 2500 issues, no nextPageToken (end of results)
        second_page_response = {
            "maxResults": 5000,
            "issues": [{"id": str(i)} for i in range(5001, 7501)]  # 2500 issues
        }
        
        call_count = 0
        def jql_request_callback(request, context):
            nonlocal call_count
            call_count += 1
            request_data = json.loads(request.text)
            
            if call_count == 1:
                # First call should have no nextPageToken
                assert "nextPageToken" not in request_data
                context.status_code = 200
                return json.dumps(first_page_response)
            else:
                # Second call should have nextPageToken
                assert request_data["nextPageToken"] == "eyJzdGFydEF0Ijo1MDAwfQ=="
                context.status_code = 200
                return json.dumps(second_page_response)
        
        mocker.register_uri(
            'POST',
            'https://test-co.atlassian.net/rest/api/2/search/jql',
            text=jql_request_callback
        )
        
        jira_connection = get_jira_mock_connection(mocker)
        result = _get_exact_count_via_pagination(jira_connection, "project = TEST")
        
        assert result == 7500  # 5000 + 2500
        assert call_count == 2  # Verify both pagination calls occurred


def test_get_exact_count_via_pagination_uses_batch_size_function():
    """Test that _get_exact_count_via_pagination uses _get_batch_size_with_jql_enhanced_search"""
    with requests_mock.Mocker() as mocker:
        # Mock response
        mock_response = {
            "maxResults": 5000,
            "issues": [{"id": "1"}]
        }
        
        def jql_request_callback(request, context):
            request_data = json.loads(request.text)
            # Verify that maxResults matches what _get_batch_size_with_jql_enhanced_search returns for ['id']
            # According to our implementation, it should return 5000 for ID-only queries
            assert request_data["maxResults"] == 5000
            assert request_data["fields"] == ["id"]
            context.status_code = 200
            return json.dumps(mock_response)
        
        mocker.register_uri(
            'POST',
            'https://test-co.atlassian.net/rest/api/2/search/jql',
            text=jql_request_callback
        )
        
        jira_connection = get_jira_mock_connection(mocker)
        result = _get_exact_count_via_pagination(jira_connection, "project = TEST")
        
        assert result == 1


def test_get_issue_count_for_jql_with_legacy_api():
    """Test that _get_issue_count_for_jql works correctly with legacy API (use_jql_enhanced_search=False)"""
    pull_from = datetime.min
    PROJECT_KEY = "PROJ"
    PROJECT_ISSUE_COUNT = 5123
    project_key_to_count = {PROJECT_KEY: PROJECT_ISSUE_COUNT}

    with requests_mock.Mocker() as mocker:
        _mock_jira_issue_by_date_endpoints(
            m=mocker,
            project_keys_to_issue_counts=project_key_to_count,
            pull_from=pull_from,
            batch_size=Constants.MAX_ISSUE_API_BATCH_SIZE,
        )
        count_for_jql = _get_issue_count_for_jql(
            get_jira_mock_connection(mocker),
            jql_query=generate_project_pull_from_jql(project_key=PROJECT_KEY, pull_from=pull_from),
            use_jql_enhanced_search=False,  # Explicitly test legacy API
        )
        assert count_for_jql == project_key_to_count[PROJECT_KEY]


def test_get_issue_count_for_jql_with_jql_enhanced_search():
    """Test that _get_issue_count_for_jql works correctly with JQL Enhanced Search API (use_jql_enhanced_search=True)"""
    PROJECT_KEY = "PROJ"
    PROJECT_ISSUE_COUNT = 2500
    jql_query = f"project = {PROJECT_KEY}"

    with requests_mock.Mocker() as mocker:
        # Mock the approximate count endpoint (now uses POST)
        mocker.register_uri(
            'POST',
            'https://test-co.atlassian.net/rest/api/2/search/approximate-count',
            json={'count': PROJECT_ISSUE_COUNT}
        )
        
        count_for_jql = _get_issue_count_for_jql(
            get_jira_mock_connection(mocker),
            jql_query=jql_query,
            use_jql_enhanced_search=True,  # Test JQL Enhanced Search API
        )
        assert count_for_jql == PROJECT_ISSUE_COUNT


def test_get_issue_count_for_jql_jql_enhanced_search_fallback_to_pagination():
    """Test that _get_issue_count_for_jql falls back to pagination when approximate count fails"""
    PROJECT_KEY = "PROJ"
    PROJECT_ISSUE_COUNT = 1500
    jql_query = f"project = {PROJECT_KEY}"

    with requests_mock.Mocker() as mocker:
        # Mock approximate count endpoint to fail
        mocker.register_uri(
            'GET',
            'https://test-co.atlassian.net/rest/api/2/search/approximate-count',
            status_code=404
        )
        
        # Mock JQL Enhanced Search endpoint for pagination fallback
        def jql_request_callback(request, context):
            request_data = json.loads(request.text)
            max_results = request_data.get('maxResults', 5000)
            
            # Return first page with issues
            return json.dumps({
                'issues': [{'id': str(i)} for i in range(1, min(max_results + 1, PROJECT_ISSUE_COUNT + 1))],
                'maxResults': max_results
                # No nextPageToken means this is the only page
            })
        
        mocker.register_uri(
            'POST',
            'https://test-co.atlassian.net/rest/api/2/search/jql',
            text=jql_request_callback
        )
        
        count_for_jql = _get_issue_count_for_jql(
            get_jira_mock_connection(mocker),
            jql_query=jql_query,
            use_jql_enhanced_search=True,  # Test JQL Enhanced Search API with fallback
        )
        assert count_for_jql == PROJECT_ISSUE_COUNT


def test_get_issue_count_for_jql_backward_compatibility():
    """Test that _get_issue_count_for_jql maintains backward compatibility when called without use_jql_enhanced_search parameter"""
    pull_from = datetime.min
    PROJECT_KEY = "PROJ"
    PROJECT_ISSUE_COUNT = 3456
    project_key_to_count = {PROJECT_KEY: PROJECT_ISSUE_COUNT}

    with requests_mock.Mocker() as mocker:
        _mock_jira_issue_by_date_endpoints(
            m=mocker,
            project_keys_to_issue_counts=project_key_to_count,
            pull_from=pull_from,
            batch_size=Constants.MAX_ISSUE_API_BATCH_SIZE,
        )
        # Call without use_jql_enhanced_search parameter - should default to False (legacy API)
        count_for_jql = _get_issue_count_for_jql(
            get_jira_mock_connection(mocker),
            jql_query=generate_project_pull_from_jql(project_key=PROJECT_KEY, pull_from=pull_from),
        )
        assert count_for_jql == project_key_to_count[PROJECT_KEY]


def test_get_issue_count_for_jql_error_handling_with_jql_enhanced_search():
    """Test that _get_issue_count_for_jql properly handles errors when using JQL Enhanced Search API"""
    jql_query = "project = TEST"
    
    # Test that JIRAError from helper function is properly propagated
    with patch(
        "jf_ingest.jf_jira.downloaders._get_issue_count_with_jql_enhanced_search",
        side_effect=JIRAError(status_code=500),
    ):
        with pytest.raises(JIRAError):
            _get_issue_count_for_jql(
                get_jira_mock_connection(), 
                jql_query=jql_query,
                use_jql_enhanced_search=True
            )


def test_get_issues_with_post_enhanced_success():
    """Test that get_issues_with_post_enhanced successfully calls the JQL Enhanced Search API"""
    with requests_mock.Mocker() as mocker:
        # Mock successful response from /search/jql endpoint
        mock_response = {
            "issues": [
                {"id": "1", "key": "TEST-1"},
                {"id": "2", "key": "TEST-2"}
            ],
            "nextPageToken": "eyJzdGFydEF0IjoyMDB9"
        }
        mocker.register_uri(
            'POST',
            'https://test-co.atlassian.net/rest/api/2/search/jql',
            json=mock_response,
            status_code=200
        )
        
        jira_connection = get_jira_mock_connection(mocker)
        issues, total, next_token = get_issues_with_post_enhanced(
            jira_connection=jira_connection,
            jql_query="project = TEST",
            fields=["id", "key"],
            expand=[],
            max_results=100
        )
        
        assert len(issues) == 2
        assert total == 2  # Should be len(issues)
        assert next_token == "eyJzdGFydEF0IjoyMDB9"
        assert issues[0]["id"] == "1"
        assert issues[1]["key"] == "TEST-2"


def test_get_issues_with_post_enhanced_with_next_page_token():
    """Test that get_issues_with_post_enhanced correctly passes nextPageToken to the API"""
    with requests_mock.Mocker() as mocker:
        # Mock successful response from /search/jql endpoint
        mock_response = {
            "issues": [
                {"id": "3", "key": "TEST-3"}
            ]
            # No nextPageToken - indicates last page
        }
        mocker.register_uri(
            'POST',
            'https://test-co.atlassian.net/rest/api/2/search/jql',
            json=mock_response,
            status_code=200
        )
        
        jira_connection = get_jira_mock_connection(mocker)
        issues, total, next_token = get_issues_with_post_enhanced(
            jira_connection=jira_connection,
            jql_query="project = TEST",
            fields=["id", "key"],
            expand=[],
            max_results=100,
            next_page_token="eyJzdGFydEF0IjoyMDB9"
        )
        
        # Verify the request included the nextPageToken
        request_data = json.loads(mocker.last_request.text)
        assert request_data["nextPageToken"] == "eyJzdGFydEF0IjoyMDB9"
        
        assert len(issues) == 1
        assert total == 1
        assert next_token is None  # No more pages
        assert issues[0]["id"] == "3"


def test_get_issues_with_post_enhanced_empty_response():
    """Test that get_issues_with_post_enhanced handles empty responses correctly"""
    with requests_mock.Mocker() as mocker:
        # Mock empty response from /search/jql endpoint
        mock_response = {
            "issues": []
            # No nextPageToken - indicates no more results
        }
        mocker.register_uri(
            'POST',
            'https://test-co.atlassian.net/rest/api/2/search/jql',
            json=mock_response,
            status_code=200
        )
        
        jira_connection = get_jira_mock_connection(mocker)
        issues, total, next_token = get_issues_with_post_enhanced(
            jira_connection=jira_connection,
            jql_query="project = EMPTY",
            fields=["id"],
            expand=[],
            max_results=100
        )
        
        assert len(issues) == 0
        assert total == 0
        assert next_token is None


def test_get_issues_with_post_enhanced_consistent_return_signature():
    """Test that get_issues_with_post_enhanced always returns exactly 3 values"""
    with requests_mock.Mocker() as mocker:
        # Mock response with various scenarios
        mock_response = {
            "issues": [{"id": "1", "key": "TEST-1"}],
            "nextPageToken": "token123"
        }
        mocker.register_uri(
            'POST',
            'https://test-co.atlassian.net/rest/api/2/search/jql',
            json=mock_response,
            status_code=200
        )
        
        jira_connection = get_jira_mock_connection(mocker)
        result = get_issues_with_post_enhanced(
            jira_connection=jira_connection,
            jql_query="project = TEST",
            fields=["id", "key"],
            expand=[],
            max_results=100
        )
        
        # Verify it always returns exactly 3 values
        assert isinstance(result, tuple)
        assert len(result) == 3
        issues, total, next_token = result
        assert isinstance(issues, list)
        assert isinstance(total, int)
        assert (next_token is None) or isinstance(next_token, str)


def test_download_issue_page_with_legacy_api():
    """Test that _download_issue_page works correctly with legacy API (use_jql_enhanced_search=False)"""
    with requests_mock.Mocker() as mocker:
        # Mock legacy API response
        mock_response = {
            "expand": "names,schema",
            "startAt": 0,
            "maxResults": 100,
            "total": 1,
            "issues": [{"id": "1", "key": "TEST-1"}]
        }
        jira_mock_connection = get_jira_mock_connection(mocker)
        _register_jira_uri(
            mocker,
            endpoint="search",
            return_value=json.dumps(mock_response),
            HTTP_ACTION='POST',
        )

        # Test with use_jql_enhanced_search=False (default)
        result = _download_issue_page(
            jira_connection=jira_mock_connection,
            jql_query="project = TEST",
            batch_size=100,
            start_at=0,
            return_total=True,
            use_jql_enhanced_search=False
        )

        # Should return 2 values for legacy API
        assert isinstance(result, tuple)
        assert len(result) == 2
        issues, total = result
        assert len(issues) == 1
        assert total == 1
        assert issues[0]["key"] == "TEST-1"


def test_download_issue_page_with_jql_enhanced_search():
    """Test that _download_issue_page works correctly with JQL Enhanced Search API (use_jql_enhanced_search=True)"""
    with requests_mock.Mocker() as mocker:
        # Mock JQL Enhanced Search API response
        mock_response = {
            "issues": [{"id": "1", "key": "TEST-1"}],
            "nextPageToken": "token123"
        }
        jira_mock_connection = get_jira_mock_connection(mocker)
        _register_jira_uri(
            mocker,
            endpoint="search/jql",
            return_value=json.dumps(mock_response),
            HTTP_ACTION='POST',
        )

        # Test with use_jql_enhanced_search=True
        result = _download_issue_page(
            jira_connection=jira_mock_connection,
            jql_query="project = TEST",
            batch_size=100,
            start_at=0,
            return_total=True,
            use_jql_enhanced_search=True,
            next_page_token=None
        )

        # Should return (issues, total) tuple when return_total=True
        assert isinstance(result, tuple)
        assert len(result) == 2
        issues, total = result
        assert len(issues) == 1
        assert total == 1  # len(issues) for JQL Enhanced Search
        assert issues[0]["key"] == "TEST-1"


def test_download_issue_page_with_jql_enhanced_search_no_return_total():
    """Test that _download_issue_page returns correct signature when return_total=False with JQL Enhanced Search"""
    with requests_mock.Mocker() as mocker:
        # Mock JQL Enhanced Search API response
        mock_response = {
            "issues": [{"id": "1", "key": "TEST-1"}],
            "nextPageToken": "token123"
        }
        jira_mock_connection = get_jira_mock_connection(mocker)
        _register_jira_uri(
            mocker,
            endpoint="search/jql",
            return_value=json.dumps(mock_response),
            HTTP_ACTION='POST',
        )

        # Test with use_jql_enhanced_search=True and return_total=False
        result = _download_issue_page(
            jira_connection=jira_mock_connection,
            jql_query="project = TEST",
            batch_size=100,
            start_at=0,
            return_total=False,
            use_jql_enhanced_search=True,
            next_page_token=None,
        )

        # Should return just issues list when return_total=False
        assert isinstance(result, list)
        assert len(result) == 1
        assert result[0]["key"] == "TEST-1"


def test_download_issue_page_backward_compatibility():
    """Test that _download_issue_page maintains backward compatibility when called without new parameters"""
    with requests_mock.Mocker() as mocker:
        # Mock legacy API response
        mock_response = {
            "expand": "names,schema",
            "startAt": 0,
            "maxResults": 100,
            "total": 1,
            "issues": [{"id": "1", "key": "TEST-1"}]
        }
        jira_mock_connection = get_jira_mock_connection(mocker)
        _register_jira_uri(
            mocker,
            endpoint="search",
            return_value=json.dumps(mock_response),
            HTTP_ACTION='POST',
        )

        # Test calling without new parameters (should use legacy API)
        result = _download_issue_page(
            jira_connection=jira_mock_connection,
            jql_query="project = TEST",
            batch_size=100,
            start_at=0,
        )

        # Should return just issues list (no total) for backward compatibility
        assert isinstance(result, list)
        assert len(result) == 1
        assert result[0]["key"] == "TEST-1"


def test_download_issue_page_error_handling_with_jql_enhanced_search():
    """Test that _download_issue_page handles errors correctly with JQL Enhanced Search API"""
    with patch("jf_ingest.jf_jira.downloaders.get_issues_with_post_enhanced", side_effect=Exception("API Error")):
        # Test error handling with JQL Enhanced Search
        result = _download_issue_page(
            jira_connection=get_jira_mock_connection(),
            jql_query="project = TEST",
            batch_size=100,
            start_at=1000,
            return_total=True,
            use_jql_enhanced_search=True,
        )

        # Should return empty results with correct signature (issues, total) when return_total=True
        assert isinstance(result, tuple)
        assert len(result) == 2
        issues, total = result
        assert issues == []
        assert total == 0


def test_get_jira_results_looped_with_legacy_api():
    """Test that get_jira_results_looped works correctly with legacy API (default behavior)"""
    with patch("jf_ingest.jf_jira.downloaders._download_issue_page") as mock_download:
        jira_connection = get_jira_mock_connection()
        jql_query = "project = TEST"
        
        # Mock two calls to _download_issue_page with 2-value returns (legacy API)
        mock_download.side_effect = [
            # First call - returns issues with total
            ([{"id": "1", "key": "TEST-1"}, {"id": "2", "key": "TEST-2"}], 3),
            # Second call - returns final issue with total
            ([{"id": "3", "key": "TEST-3"}], 3)
        ]
        
        results = get_jira_results_looped(
            jira_connection=jira_connection,
            jql_query=jql_query,
            batch_size=2,
            issue_count=3,
            include_fields=[
                JiraFieldIdentifier(jira_field_id='id', jira_field_name='id'),
                JiraFieldIdentifier(jira_field_id='key', jira_field_name='key'),
            ]
        )
        
        assert len(results) == 3
        assert results[0]["id"] == "1"
        assert results[1]["id"] == "2"
        assert results[2]["id"] == "3"
        
        # Verify _download_issue_page was called correctly for legacy API
        assert mock_download.call_count == 2
        
        # First call should have start_at=0
        first_call = mock_download.call_args_list[0]
        assert first_call[1]["start_at"] == 0
        assert "use_jql_enhanced_search" not in first_call[1] or first_call[1]["use_jql_enhanced_search"] is False
        
        # Second call should have start_at=2 (after processing 2 results)
        second_call = mock_download.call_args_list[1]
        assert second_call[1]["start_at"] == 2
        assert "use_jql_enhanced_search" not in second_call[1] or second_call[1]["use_jql_enhanced_search"] is False


def test_get_jira_results_looped_with_jql_enhanced_search():
    """Test that get_jira_results_looped works correctly with JQL Enhanced Search API"""
    with patch("jf_ingest.jf_jira.downloaders._download_issue_page") as mock_download:
        jira_connection = get_jira_mock_connection()
        jql_query = "project = TEST"
        
        # For this test, let's use a simple single-page scenario to avoid the complexity
        # Mock single call that returns all results with next_page_token=None (no more pages)
        mock_download.return_value = ([{"id": "1", "key": "TEST-1"}, {"id": "2", "key": "TEST-2"}, {"id": "3", "key": "TEST-3"}], None)
        
        results = get_jira_results_looped(
            jira_connection=jira_connection,
            jql_query=jql_query,
            batch_size=10,
            issue_count=3,
            include_fields=[
                JiraFieldIdentifier(jira_field_id='id', jira_field_name='id'),
                JiraFieldIdentifier(jira_field_id='key', jira_field_name='key'),
            ],
            use_jql_enhanced_search=True
        )
        
        assert len(results) == 3
        assert results[0]["id"] == "1"
        assert results[1]["id"] == "2"
        assert results[2]["id"] == "3"
        
        # Verify _download_issue_page was called correctly
        assert mock_download.call_count == 1
        
        # Call should have use_jql_enhanced_search=True and next_page_token=None (first call)
        call_args = mock_download.call_args_list[0]
        assert call_args[1]["use_jql_enhanced_search"] is True
        assert call_args[1]["next_page_token"] is None


def test_get_jira_results_looped_backward_compatibility():
    """Test that get_jira_results_looped maintains backward compatibility when use_jql_enhanced_search is not provided"""
    with patch("jf_ingest.jf_jira.downloaders._download_issue_page") as mock_download:
        jira_connection = get_jira_mock_connection()
        jql_query = "project = TEST"
        
        # Mock single call to _download_issue_page
        mock_download.return_value = ([{"id": "1", "key": "TEST-1"}], 1)
        
        # Call without use_jql_enhanced_search parameter (should default to False)
        results = get_jira_results_looped(
            jira_connection=jira_connection,
            jql_query=jql_query,
            batch_size=10,
            issue_count=1
        )
        
        assert len(results) == 1
        assert results[0]["id"] == "1"
        
        # Verify _download_issue_page was called with legacy API behavior
        assert mock_download.call_count == 1
        call_args = mock_download.call_args_list[0]
        assert "use_jql_enhanced_search" not in call_args[1] or call_args[1]["use_jql_enhanced_search"] is False


def test_pull_jira_issues_by_jira_ids_with_use_jql_enhanced_search_parameter():
    """Test that pull_jira_issues_by_jira_ids passes use_jql_enhanced_search parameter correctly"""
    with patch("jf_ingest.jf_jira.downloaders._download_issue_page") as mock_download:
        jira_connection = get_jira_mock_connection()
        jira_ids = ["1", "2"]
        
        # Mock _download_issue_page to return issues
        mock_download.return_value = [
            {"id": "1", "key": "TEST-1"},
            {"id": "2", "key": "TEST-2"}
        ]
        
        # Test with use_jql_enhanced_search=True
        list(pull_jira_issues_by_jira_ids(
            jira_connection=jira_connection,
            jira_ids=jira_ids,
            num_parallel_threads=1,
            batch_size=10,
            use_jql_enhanced_search=True
        ))
        
        # Verify _download_issue_page was called with use_jql_enhanced_search=True
        assert mock_download.call_count == 1
        call_args = mock_download.call_args_list[0]
        assert call_args[1]["use_jql_enhanced_search"] is True


def test_pull_jira_issues_by_jira_ids_backward_compatibility():
    """Test that pull_jira_issues_by_jira_ids maintains backward compatibility when use_jql_enhanced_search is not provided"""
    with patch("jf_ingest.jf_jira.downloaders._download_issue_page") as mock_download:
        jira_connection = get_jira_mock_connection()
        jira_ids = ["1"]
        
        # Mock _download_issue_page to return issues
        mock_download.return_value = [{"id": "1", "key": "TEST-1"}]
        
        # Test without use_jql_enhanced_search parameter (should default to False)
        list(pull_jira_issues_by_jira_ids(
            jira_connection=jira_connection,
            jira_ids=jira_ids,
            num_parallel_threads=1,
            batch_size=10
        ))
        
        # Verify _download_issue_page was called with use_jql_enhanced_search=False (default)
        assert mock_download.call_count == 1
        call_args = mock_download.call_args_list[0]
        # The parameter should not be present in the call (using default)
        assert "use_jql_enhanced_search" not in call_args[1] or call_args[1]["use_jql_enhanced_search"] is False


def test_pull_all_jira_issues_by_date_with_use_jql_enhanced_search_parameter():
    """Test that pull_all_jira_issues_by_date passes use_jql_enhanced_search parameter correctly"""
    from datetime import datetime
    
    with patch("jf_ingest.jf_jira.downloaders._get_all_project_issue_counts") as mock_count, \
         patch("jf_ingest.jf_jira.downloaders._download_issue_page") as mock_download, \
         patch("jf_ingest.jf_jira.downloaders._pull_all_jira_issues_by_date_enhanced") as mock_enhanced:
        
        jira_connection = get_jira_mock_connection()
        project_key_to_pull_from = {"TEST": datetime.min}
        
        # Mock _get_all_project_issue_counts to return project counts
        mock_count.return_value = {"TEST": 2}
        
        # Mock _pull_all_jira_issues_by_date_enhanced to return issues
        mock_enhanced.return_value = iter([
            {"id": "1", "key": "TEST-1"},
            {"id": "2", "key": "TEST-2"}
        ])
        
        # Mock _download_issue_page to return issues in the correct format for JQL Enhanced Search
        # Format: (issues_list, total_count, next_page_token)
        mock_download.return_value = (
            [{"id": "1", "key": "TEST-1"}, {"id": "2", "key": "TEST-2"}],  # issues
            2,  # total
            None  # next_page_token (no more pages)
        )
        
        # Test with use_jql_enhanced_search=True
        list(pull_all_jira_issues_by_date(
            jira_connection=jira_connection,
            project_key_to_pull_from=project_key_to_pull_from,
            num_parallel_threads=1,
            batch_size=10,
            jql_project_batch_size=1,
            use_jql_enhanced_search=True
        ))
        
        # Verify _get_all_project_issue_counts was called with use_jql_enhanced_search=True
        assert mock_count.call_count == 1
        count_call_args = mock_count.call_args_list[0]
        assert count_call_args[1]["use_jql_enhanced_search"] is True
        
        # Verify _pull_all_jira_issues_by_date_enhanced was called (JQL Enhanced Search path)
        assert mock_enhanced.call_count == 1
        enhanced_call_args = mock_enhanced.call_args_list[0]
        # Verify the enhanced function was called with the correct parameters
        assert enhanced_call_args[1]["jira_connection"] == jira_connection
        assert enhanced_call_args[1]["project_key_to_pull_from"] == project_key_to_pull_from


def test_pull_all_jira_issues_by_date_backward_compatibility():
    """Test that pull_all_jira_issues_by_date maintains backward compatibility when use_jql_enhanced_search is not provided"""
    from datetime import datetime
    
    with patch("jf_ingest.jf_jira.downloaders._get_all_project_issue_counts") as mock_count, \
         patch("jf_ingest.jf_jira.downloaders._download_issue_page") as mock_download:
        
        jira_connection = get_jira_mock_connection()
        project_key_to_pull_from = {"TEST": datetime.min}
        
        # Mock _get_all_project_issue_counts to return project counts
        mock_count.return_value = {"TEST": 1}
        
        # Mock _download_issue_page to return issues
        mock_download.return_value = [{"id": "1", "key": "TEST-1"}]
        
        # Test without use_jql_enhanced_search parameter (should default to False)
        list(pull_all_jira_issues_by_date(
            jira_connection=jira_connection,
            project_key_to_pull_from=project_key_to_pull_from,
            num_parallel_threads=1,
            batch_size=10,
            jql_project_batch_size=1
        ))
        
        # Verify _get_all_project_issue_counts was called with use_jql_enhanced_search=False (default)
        assert mock_count.call_count == 1
        count_call_args = mock_count.call_args_list[0]
        assert "use_jql_enhanced_search" not in count_call_args[1] or count_call_args[1]["use_jql_enhanced_search"] is False
        
        # Verify _download_issue_page was called with use_jql_enhanced_search=False (default)
        assert mock_download.call_count >= 1
        download_call_args = mock_download.call_args_list[0]
        assert "use_jql_enhanced_search" not in download_call_args[1] or download_call_args[1]["use_jql_enhanced_search"] is False


def test_get_all_project_issue_counts_with_use_jql_enhanced_search_parameter():
    """Test that _get_all_project_issue_counts passes use_jql_enhanced_search parameter correctly"""
    from datetime import datetime
    
    with patch("jf_ingest.jf_jira.downloaders._get_issue_count_for_jql") as mock_count:
        jira_connection = get_jira_mock_connection()
        project_key_to_pull_from = {"TEST": datetime.min}
        
        # Mock _get_issue_count_for_jql to return a count
        mock_count.return_value = 5
        
        # Test with use_jql_enhanced_search=True
        _get_all_project_issue_counts(
            jira_connection=jira_connection,
            project_key_to_pull_from=project_key_to_pull_from,
            num_parallel_threads=1,
            use_jql_enhanced_search=True
        )
        
        # Verify _get_issue_count_for_jql was called with use_jql_enhanced_search=True
        assert mock_count.call_count == 1
        call_args = mock_count.call_args_list[0]
        assert call_args[1]["use_jql_enhanced_search"] is True


def test_get_all_project_issue_counts_backward_compatibility():
    """Test that _get_all_project_issue_counts maintains backward compatibility when use_jql_enhanced_search is not provided"""
    from datetime import datetime
    
    with patch("jf_ingest.jf_jira.downloaders._get_issue_count_for_jql") as mock_count:
        jira_connection = get_jira_mock_connection()
        project_key_to_pull_from = {"TEST": datetime.min}
        
        # Mock _get_issue_count_for_jql to return a count
        mock_count.return_value = 3
        
        # Test without use_jql_enhanced_search parameter (should default to False)
        _get_all_project_issue_counts(
            jira_connection=jira_connection,
            project_key_to_pull_from=project_key_to_pull_from,
            num_parallel_threads=1
        )
        
        # Verify _get_issue_count_for_jql was called with use_jql_enhanced_search=False (default)
        assert mock_count.call_count == 1
        call_args = mock_count.call_args_list[0]
        assert "use_jql_enhanced_search" not in call_args[1] or call_args[1]["use_jql_enhanced_search"] is False


def test_fetch_id_to_key_for_all_existing_parameter_passing():
    """Test that fetch_id_to_key_for_all_existing passes use_jql_enhanced_search parameter through to called functions."""
    from unittest.mock import patch, MagicMock
    
    # Mock the functions that fetch_id_to_key_for_all_existing calls
    with patch('jf_ingest.jf_jira.downloaders._get_all_project_issue_counts') as mock_get_counts, \
         patch('jf_ingest.jf_jira.downloaders.get_jira_search_batch_size') as mock_batch_size, \
         patch('jf_ingest.jf_jira.downloaders.get_jira_results_looped') as mock_results_looped, \
         patch('jf_ingest.jf_jira.downloaders.generate_project_pull_from_jql') as mock_jql:
        
        # Set up mock return values
        mock_get_counts.return_value = {'PROJ': 100}
        mock_batch_size.return_value = 1000
        mock_results_looped.return_value = [{'id': '1', 'key': 'PROJ-1'}]
        mock_jql.return_value = 'project = PROJ'
        
        # Test with use_jql_enhanced_search=True
        result = fetch_id_to_key_for_all_existing(
            jira_connection=MagicMock(),
            project_ids=['PROJ'],
            pull_from=datetime.min,
            jql_filter=None,
            use_jql_enhanced_search=True
        )
        
        # Verify the parameter was passed to _get_all_project_issue_counts
        mock_get_counts.assert_called_once()
        call_args = mock_get_counts.call_args
        assert call_args[1]['use_jql_enhanced_search'] is True
        
        # Verify the parameter was passed to get_jira_results_looped
        mock_results_looped.assert_called_once()
        call_args = mock_results_looped.call_args
        assert call_args[1]['use_jql_enhanced_search'] is True
        
        # Verify return value
        assert result == {'1': 'PROJ-1'}


def test_fetch_id_to_key_for_all_existing_backward_compatibility():
    """Test that fetch_id_to_key_for_all_existing maintains backward compatibility when use_jql_enhanced_search is not provided."""
    from unittest.mock import patch, MagicMock
    
    # Mock the functions that fetch_id_to_key_for_all_existing calls
    with patch('jf_ingest.jf_jira.downloaders._get_all_project_issue_counts') as mock_get_counts, \
         patch('jf_ingest.jf_jira.downloaders.get_jira_search_batch_size') as mock_batch_size, \
         patch('jf_ingest.jf_jira.downloaders.get_jira_results_looped') as mock_results_looped, \
         patch('jf_ingest.jf_jira.downloaders.generate_project_pull_from_jql') as mock_jql:
        
        # Set up mock return values
        mock_get_counts.return_value = {'PROJ': 100}
        mock_batch_size.return_value = 1000
        mock_results_looped.return_value = [{'id': '1', 'key': 'PROJ-1'}]
        mock_jql.return_value = 'project = PROJ'
        
        # Test without use_jql_enhanced_search parameter (should default to False)
        result = fetch_id_to_key_for_all_existing(
            jira_connection=MagicMock(),
            project_ids=['PROJ'],
            pull_from=datetime.min,
            jql_filter=None
        )
        
        # Verify the parameter was passed as False to _get_all_project_issue_counts
        mock_get_counts.assert_called_once()
        call_args = mock_get_counts.call_args
        assert call_args[1]['use_jql_enhanced_search'] is False
        
        # Verify the parameter was passed as False to get_jira_results_looped
        mock_results_looped.assert_called_once()
        call_args = mock_results_looped.call_args
        assert call_args[1]['use_jql_enhanced_search'] is False
        
        # Verify return value
        assert result == {'1': 'PROJ-1'}


def test_fetch_id_to_key_for_all_existing_api_selection_integration():
    """Integration test to verify API selection works through the entire call chain."""
    from unittest.mock import patch, MagicMock
    
    # Mock the lower-level functions to verify they receive the correct parameter
    with patch('jf_ingest.jf_jira.downloaders._get_issue_count_for_jql') as mock_count_jql, \
         patch('jf_ingest.jf_jira.downloaders.get_jira_search_batch_size') as mock_batch_size, \
         patch('jf_ingest.jf_jira.downloaders.get_jira_results_looped') as mock_results_looped, \
         patch('jf_ingest.jf_jira.downloaders.generate_project_pull_from_jql') as mock_jql:
        
        # Set up mock return values
        mock_count_jql.return_value = 100
        mock_batch_size.return_value = 5000  # JQL Enhanced Search batch size
        mock_results_looped.return_value = [{'id': '1', 'key': 'PROJ-1'}]
        mock_jql.return_value = 'project = PROJ'
        
        # Test with use_jql_enhanced_search=True
        result = fetch_id_to_key_for_all_existing(
            jira_connection=MagicMock(),
            project_ids=['PROJ'],
            pull_from=datetime.min,
            jql_filter=None,
            use_jql_enhanced_search=True
        )
        
        # Verify _get_issue_count_for_jql was called with use_jql_enhanced_search=True
        # This happens through _get_all_project_issue_counts
        mock_count_jql.assert_called()
        count_call_args = mock_count_jql.call_args
        assert count_call_args[1]['use_jql_enhanced_search'] is True
        
        # Verify get_jira_results_looped was called with use_jql_enhanced_search=True
        mock_results_looped.assert_called_once()
        results_call_args = mock_results_looped.call_args
        assert results_call_args[1]['use_jql_enhanced_search'] is True
        
        # Verify return value
        assert result == {'1': 'PROJ-1'}



import json
import logging
from dataclasses import dataclass
from datetime import datetime, timezone
from typing import Optional, cast

from requests import Response, Session

from jf_ingest.utils import retry_for_status

logger = logging.getLogger(__name__)


GQL_PAGE_INFO_BLOCK = "pageInfo {hasNextPage, endCursor}"


@dataclass
class GQLRateLimit:
    remaining: int
    reset_at: datetime


class GqlRateLimitedExceptionInner(Exception):
    pass


def gql_format_to_datetime(datetime_str: str) -> datetime:
    """Attempt to formate a datetime str from the GQL format to a python Datetime Object
    NOTE: This is the format shared by the Github and Gitlab GQL clients

    Args:
        datetime_str (str): The datetime from graphql

    Returns:
        datetime: A valid, timezone aware datetime
    """
    return datetime.strptime(datetime_str, "%Y-%m-%dT%H:%M:%SZ").replace(tzinfo=timezone.utc)




import json
from jf_ingest.jf_jira.utils import JiraFieldIdentifier, expand_and_normalize_jira_fields
from tests.jf_jira.utils import get_fixture_file_data

def get_field_data() -> list[dict]:
    return [
    {
        "id": "statuscategorychangedate",
        "key": "statuscategorychangedate",
        "name": "Status Category Changed",
        "custom": False,
        "orderable": False,
        "navigable": True,
        "searchable": True,
        "clauseNames": [
            "statusCategoryChangedDate"
        ],
        "schema": {
            "type": "datetime",
            "system": "statuscategorychangedate"
        }
    },
    {
        "id": "parent",
        "key": "parent",
        "name": "Parent",
        "custom": False,
        "orderable": False,
        "navigable": True,
        "searchable": False,
        "clauseNames": [
            "parent"
        ]
    },
    {
        "id": "customfield_10070",
        "key": "customfield_10070",
        "name": "Satisfaction date",
        "untranslatedName": "Satisfaction date",
        "custom": True,
        "orderable": True,
        "navigable": True,
        "searchable": True,
        "clauseNames": [
            "cf[10070]",
            "Satisfaction date"
        ],
        "schema": {
            "type": "datetime",
            "custom": "com.atlassian.servicedesk:sd-request-feedback-date",
            "customId": 10070
        }
    },
    {
        "id": "customfield_10071",
        "key": "customfield_10071",
        "name": "Reporters",
        "untranslatedName": "Reporters",
        "custom": True,
        "orderable": True,
        "navigable": True,
        "searchable": True,
        "clauseNames": [
            "Reporters",
            "Reporters[User Picker (multiple users)]",
            "cf[10071]"
        ],
        "schema": {
            "type": "array",
            "items": "user",
            "custom": "com.atlassian.jira.plugin.system.customfieldtypes:multiuserpicker",
            "customId": 10071
        }
    }
]

def test_expand_and_normalize_jira_fields_smoke_test():
    raw_fields = get_field_data()
    
    jira_field_identifiers = expand_and_normalize_jira_fields(fields_from_jira=raw_fields, field_names_or_ids=[])
    assert jira_field_identifiers == []
    

def test_expand_and_normalize_jira_fields_with_id():
    raw_fields = get_field_data()
    
    fields_to_expand = ['customfield_10070']
    jira_field_identifiers = expand_and_normalize_jira_fields(fields_from_jira=raw_fields, field_names_or_ids=fields_to_expand)
    assert len(jira_field_identifiers) == 1
    assert jira_field_identifiers[0].jira_field_id == 'customfield_10070'
    assert jira_field_identifiers[0].jira_field_name == 'Satisfaction date'

def test_expand_and_normalize_jira_fields_with_name():
    raw_fields = get_field_data()
    
    fields_to_expand = ['Satisfaction date']
    jira_field_identifiers = expand_and_normalize_jira_fields(fields_from_jira=raw_fields, field_names_or_ids=fields_to_expand)
    assert len(jira_field_identifiers) == 1
    assert jira_field_identifiers[0].jira_field_id == 'customfield_10070'
    assert jira_field_identifiers[0].jira_field_name == 'Satisfaction date'

def test_expand_and_normalize_jira_fields_with_redundant_data():
    raw_fields = get_field_data()
    
    fields_to_expand = ['customfield_10070', 'Satisfaction date']
    jira_field_identifiers = expand_and_normalize_jira_fields(fields_from_jira=raw_fields, field_names_or_ids=fields_to_expand)
    assert len(jira_field_identifiers) == 1
    assert jira_field_identifiers[0].jira_field_id == 'customfield_10070'
    assert jira_field_identifiers[0].jira_field_name == 'Satisfaction date'
    
def test_expand_and_normalize_jira_fields_with_all_ids():
    raw_fields = sorted(get_field_data(), key=lambda _dict: _dict['id'])
    ids = [f['id'] for f in raw_fields]
    names = [f['name'] for f in raw_fields]
    assert len(ids) == len(names)
    
    
    jira_field_identifiers = expand_and_normalize_jira_fields(fields_from_jira=raw_fields, field_names_or_ids=ids)
    
    assert len(jira_field_identifiers) == len(ids)
    for id, name, jira_field_identifier in zip(ids, names, sorted(jira_field_identifiers, key=lambda jfi: jfi.jira_field_id)):
        print(id, name, jira_field_identifier)
        assert jira_field_identifier.jira_field_id == id
        assert jira_field_identifier.jira_field_name == name

def test_expand_and_normalize_jira_fields_with_all_names():
    raw_fields = sorted(get_field_data(), key=lambda _dict: _dict['id'])
    ids = [f['id'] for f in raw_fields]
    names = [f['name'] for f in raw_fields]
    assert len(ids) == len(names)
    
    
    jira_field_identifiers = expand_and_normalize_jira_fields(fields_from_jira=raw_fields, field_names_or_ids=names)
    
    assert len(jira_field_identifiers) == len(names)
    for id, name, jira_field_identifier in zip(ids, names, sorted(jira_field_identifiers, key=lambda jfi: jfi.jira_field_id)):
        print(id, name, jira_field_identifier)
        assert jira_field_identifier.jira_field_id == id
        assert jira_field_identifier.jira_field_name == name

def test_expand_and_normalize_jira_fields_with_missing_field():
    """This test specifically tests if you include a field that doesn't exist
    in the "raw_fields" data (returned by the field API endpoint), we will still
    create a JiraFieldIdentifier for the include field. In that use case, we will
    set the provided include_field as both the ID and the Name.
    This is a specific fix for some weird versions of Jira Server, where the
    parent field is not returned by the Jira Fields API endpoint, but we do still
    expect to see the parent field on the JiraIssue objects
    """
    # Set up raw fields WITH NO parent data
    raw_fields = [
        {
            "id": "customfield_10070",
            "key": "customfield_10070",
            "name": "Satisfaction date",
            "untranslatedName": "Satisfaction date",
            "custom": True,
            "orderable": True,
            "navigable": True,
            "searchable": True,
            "clauseNames": [
                "cf[10070]",
                "Satisfaction date"
            ],
            "schema": {
                "type": "datetime",
                "custom": "com.atlassian.servicedesk:sd-request-feedback-date",
                "customId": 10070
            }
        },
        {
            "id": "customfield_10071",
            "key": "customfield_10071",
            "name": "Reporters",
            "untranslatedName": "Reporters",
            "custom": True,
            "orderable": True,
            "navigable": True,
            "searchable": True,
            "clauseNames": [
                "Reporters",
                "Reporters[User Picker (multiple users)]",
                "cf[10071]"
            ],
            "schema": {
                "type": "array",
                "items": "user",
                "custom": "com.atlassian.jira.plugin.system.customfieldtypes:multiuserpicker",
                "customId": 10071
            }
        }
    ]
    raw_fields = sorted(raw_fields, key=lambda _dict: _dict['id'])
    
    jira_field_identifiers = expand_and_normalize_jira_fields(fields_from_jira=raw_fields, field_names_or_ids=['parent', 'id', 'customfield_10070'])
    assert len(jira_field_identifiers) == 3
    jira_field_identifiers_ids = [jfi.jira_field_id for jfi in jira_field_identifiers]
    jira_field_identifiers_names = [jfi.jira_field_name for jfi in jira_field_identifiers]
    
    # Verify IDs
    assert 'parent' in jira_field_identifiers_ids
    assert 'id' in jira_field_identifiers_ids
    assert 'customfield_10070' in jira_field_identifiers_ids
    assert 'customfield_10071' not in jira_field_identifiers_ids
    
    # Verify Names
    assert 'parent' in jira_field_identifiers_names
    assert 'id' in jira_field_identifiers_names
    assert 'Satisfaction date' in jira_field_identifiers_names
    assert 'Reporters' not in jira_field_identifiers_names
    
        
def test_matches_changelog_item_base():
    field_name = 'FieldName'
    field_id = 'customfield_xxxxx'
    changelog_item = {
        'field': field_name,
        'fieldId': field_id
    }
    
    jfi = JiraFieldIdentifier(jira_field_id=field_id, jira_field_name=field_name)
    assert jfi.matches_changelog_item(changelog_item)
    
    changelog_item_no_id = {
        'field': field_name
    }
    
    assert jfi.matches_changelog_item(changelog_item_no_id)
    
    changelog_item_no_name = {
        'fieldId': field_id
    }
    assert jfi.matches_changelog_item(changelog_item_no_name)
    
    changelog_item_reversed = {
        'fieldId': field_name,
        'field': field_id,
    }
    assert not jfi.matches_changelog_item(changelog_item_reversed)
    
    changelog_item_reversed = {
        'fieldId': 'nomatch',
        'field': 'nomatch',
    }
    assert not jfi.matches_changelog_item(changelog_item_reversed)
    

        
def test_matches_changelog_capitalization_test():
    field_name = 'status'
    field_id = 'status'
    
    jfi = JiraFieldIdentifier(jira_field_id=field_id, jira_field_name=field_name)
    assert jfi.matches_changelog_item({
        'field': field_name
    })
    
    # This should not match, because this test should be case sensitive
    assert not jfi.matches_changelog_item({
        'field': 'Status'
    })
    
    # This should match because Jira type fields are not case sensitive
    assert jfi.matches_changelog_item({
        'field': 'Status',
        'fieldtype': 'jira'
    })
    
    
    # This should not match because non-Jira type fields are case sensitive
    assert not jfi.matches_changelog_item({
        'field': 'Status',
        'fieldtype': 'custom'
    })
